import logging
from collections.abc import Callable
from fnmatch import (
    fnmatch,
)

import reactivex
from pydantic import (
    BaseModel,
    ConfigDict,
)
from reactivex import (
    Observable,
)
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)
from reactivex.scheduler import (
    ThreadPoolScheduler,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.alpine.parse_apk_db import (
    parse_apk_db,
)
from labels.pkg.cataloger.arch.parse_alpm import (
    parse_alpm_db,
)
from labels.pkg.cataloger.debian.parse_dpkg_db import (
    parse_dpkg_db,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
    Parser,
)
from labels.pkg.cataloger.redhat.parse_rpm_db import (
    parse_rpm_db,
)

LOGGER = logging.getLogger(__name__)


class Request(BaseModel):
    real_path: str
    parser: Parser
    parser_name: str
    model_config = ConfigDict(frozen=True)


class Task(BaseModel):
    location: Location
    parser: Parser
    parser_name: str
    model_config = ConfigDict(frozen=True)


def execute_parsers(
    resolver: Resolver,
    environment: Environment,
) -> Callable[[Observable[Task]], Observable]:
    def _handle(source: Observable[Task]) -> Observable:
        def subscribe(
            observer: ObserverBase[tuple[list[Package], list[Relationship]]],
            scheduler: ThreadPoolScheduler | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Task) -> None:
                LOGGER.debug("Working on %s", value.location.access_path)
                content_reader = resolver.file_contents_by_location(value.location)
                try:
                    if content_reader is not None and (
                        result := value.parser(
                            resolver,
                            environment,
                            LocationReadCloser(
                                location=value.location,
                                read_closer=content_reader,
                            ),
                        )
                    ):
                        discover_packages, relationships = result
                        for pkg in discover_packages:
                            pkg.found_by = value.parser_name
                        observer.on_next((discover_packages, relationships))
                except Exception as ex:  # noqa: BLE001
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)  # type: ignore

    return _handle


def on_next_db_file(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if "lib/apk/db/installed" in value:
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_apk_db,
                            parser_name="apk-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/var/lib/dpkg/status",
                        "*var/lib/dpkg/status",
                        "/var/lib/dpkg/status",
                        "**/var/lib/dpkg/status.d/*",
                        "*var/lib/dpkg/status.d/*",
                        "/var/lib/dpkg/status.d/*",
                        "**/lib/opkg/info/*.control",
                        "*lib/opkg/info/*.control",
                        "/lib/opkg/info/*.control",
                        "**/lib/opkg/status",
                        "*lib/opkg/status",
                        "/lib/opkg/status",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_dpkg_db,
                            parser_name="dpkg-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern)
                    for pattern in (
                        "**/var/lib/pacman/local/**/desc",
                        "var/lib/pacman/local/**/desc",
                        "/var/lib/pacman/local/**/desc",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_alpm_db,
                            parser_name="alpm-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        (
                            "**/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        (
                            "/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        "**/rpmdb.sqlite",
                        "**/var/lib/rpm/Packages",
                        "**/var/lib/rpm/Packages.db",
                        "**/var/lib/rpm/rpmdb.sqlite",
                        "**/usr/share/rpm/Packages",
                        "**/usr/share/rpm/Packages.db",
                        "**/usr/share/rpm/rpmdb.sqlite",
                        "**/usr/lib/sysimage/rpm/Packages",
                        "**/usr/lib/sysimage/rpm/Packages.db",
                        "**/usr/lib/sysimage/rpm/rpmdb.sqlite",
                        "/var/lib/rpm/Packages",
                        "/var/lib/rpm/Packages.db",
                        "/var/lib/rpm/rpmdb.sqlite",
                        "/usr/share/rpm/Packages",
                        "/usr/share/rpm/Packages.db",
                        "/usr/share/rpm/rpmdb.sqlite",
                        "/usr/lib/sysimage/rpm/Packages",
                        "/usr/lib/sysimage/rpm/Packages.db",
                        "/usr/lib/sysimage/rpm/rpmdb.sqlite",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_rpm_db,
                            parser_name="environment-parser",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import logging
import os
from copy import (
    deepcopy,
)
from ctypes import (
    c_void_p,
    cdll,
)

from tree_sitter import (
    Language as TLanguage,
)
from tree_sitter import (
    Node,
    Parser,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.elixir.package import (
    ElixirMixLockEntry,
    new_package,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.utils.env import (
    TREE_SITTER_PARSERS,
)

LOGGER = logging.getLogger(__name__)


def lang_from_so(path: str, name: str) -> TLanguage:
    lib = cdll.LoadLibrary(os.fspath(path))
    language_function = getattr(lib, f"tree_sitter_{name}")
    language_function.restype = c_void_p
    language_ptr = language_function()
    return TLanguage(language_ptr)


def process_entry(entry: Node, location: Location) -> Package | None:
    name_value = next(x for x in entry.named_children if x.type == "package_name").text
    version_value = next(x for x in entry.named_children if x.type == "version").text
    pkg_hash_value = next(x for x in entry.named_children if x.type == "checksum").text
    pkg_hash_ext_value = next(x for x in entry.named_children if x.type == "optional_checksum").text

    if pkg_hash_ext_value and pkg_hash_value and version_value and name_value:
        name = name_value.decode("utf-8")[1:-1]
        version = version_value.decode("utf-8")[1:-1]
        pkg_hash = pkg_hash_value.decode("utf-8")[1:-1]
        pkg_hash_ext = pkg_hash_ext_value.decode("utf-8")[1:-1]
    else:
        return None

    new_location = deepcopy(location)
    if new_location.coordinates:
        new_location.coordinates.line = entry.start_point[0] + 1

    package = new_package(
        ElixirMixLockEntry(
            name=name,
            version=version,
            pkg_hash=pkg_hash,
            pkg_hash_ext=pkg_hash_ext,
        ),
        new_location,
    )

    return package if package else None


def collect_dependencies(
    packages: list[Package],
    package_entries: list[Node],
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for entry in package_entries:
        pkg_name_value = next(x for x in entry.named_children if x.type == "package_name").text
        if pkg_name_value:
            current_package_name = pkg_name_value.decode("utf-8")[1:-1]
        else:
            continue

        for dependency_list in next(
            node for node in entry.named_children if node.type == "dependencies"
        ).named_children:
            dependencies = [
                value.decode("utf-8")[1:]
                for x in dependency_list.named_children
                if x.type == "dependency"
                and (value := next(y for y in x.named_children if y.type == "atom").text)
            ]
            relationships.extend(
                Relationship(
                    from_=next(x for x in packages if x.name == dep_name),
                    to_=next(x for x in packages if x.name == current_package_name),
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for dep_name in dependencies
                if next((x for x in packages if x.name == dep_name), None)
            )
    return relationships


def parse_mix_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    if TREE_SITTER_PARSERS is None:
        LOGGER.warning(
            (
                "Unable to parse mix.lock file because the binary file "
                "for the parser is not available in the environment."
            ),
        )
        return packages, relationships

    so_library_path: str = os.path.join(TREE_SITTER_PARSERS, "mix_lock.so")
    parser_language = lang_from_so(so_library_path, "mix_lock")
    parser = Parser(parser_language)

    result = parser.parse(reader.read_closer.read().encode("utf-8"))
    package_entries = [node for node in result.root_node.children if node.type == "package_entry"]

    packages = [package for x in package_entries if (package := process_entry(x, reader.location))]
    relationships = collect_dependencies(packages, package_entries)

    return packages, relationships



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import UTC, datetime
from multiprocessing import cpu_count
from typing import TYPE_CHECKING

from spdx_tools.spdx.model.actor import Actor, ActorType
from spdx_tools.spdx.model.document import CreationInfo, Document
from spdx_tools.spdx.model.package import Package as SPDX_Package
from spdx_tools.spdx.model.package import PackagePurpose
from spdx_tools.spdx.model.relationship import Relationship as SPDXRelationship
from spdx_tools.spdx.model.relationship import RelationshipType as SPDXRelationshipType
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document

from labels.file.resolver import Resolver

if TYPE_CHECKING:
    from spdx_tools.spdx.validation.validation_message import ValidationMessage
from spdx_tools.spdx.writer.write_anything import write_file

from labels.artifact.relationship import Relationship
from labels.format.common import get_document_namespace, set_namespace_version
from labels.format.spdx.complete_file import (
    NOASSERTION,
    add_authors,
    add_empty_package,
    add_external_refs,
    add_integrity,
    add_license,
    get_spdx_id,
)
from labels.model.core import Package, SbomConfig
from labels.utils.exceptions import SPDXValidationError

LOGGER = logging.getLogger(__name__)


def package_to_spdx_pkg(package: Package) -> SPDX_Package:
    return SPDX_Package(
        spdx_id=get_spdx_id(package),
        name=package.name,
        download_location=NOASSERTION,
        version=package.version,
        license_declared=add_license(package.licenses),
        originator=add_authors(package.health_metadata),
        primary_package_purpose=PackagePurpose.LIBRARY,
        external_references=add_external_refs(package),
        checksums=add_integrity(package.health_metadata),
    )


def create_package_cache(packages: list[Package]) -> dict[Package, SPDX_Package]:
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        return dict(zip(packages, executor.map(package_to_spdx_pkg, packages), strict=False))


def create_document_relationships(
    document: Document,
    spdx_packages: list[SPDX_Package],
) -> list[SPDXRelationship]:
    doc_spdx_id = document.creation_info.spdx_id
    return [
        SPDXRelationship(doc_spdx_id, SPDXRelationshipType.DESCRIBES, pkg.spdx_id)
        for pkg in spdx_packages
    ]


def process_relationships(
    document: Document,
    _relationships: list[Relationship],
    spdx_id_cache: dict[Package, str],
    document_relationships: list[SPDXRelationship],
) -> None:
    def process_relationship(
        relationship: Relationship,
    ) -> SPDXRelationship | None:
        to_pkg = relationship.to_
        from_pkg = relationship.from_

        if isinstance(to_pkg, Package) and isinstance(from_pkg, Package):
            to_pkg_id = spdx_id_cache.get(to_pkg)
            from_pkg_id = spdx_id_cache.get(from_pkg)

            if to_pkg_id and from_pkg_id:
                return SPDXRelationship(to_pkg_id, SPDXRelationshipType.DEPENDENCY_OF, from_pkg_id)
        return None

    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        relationship_results = filter(None, executor.map(process_relationship, _relationships))
        document_relationships.extend(relationship_results)

    document.relationships = document_relationships


def add_packages_and_relationships(
    document: Document,
    packages: list[Package],
    _relationships: list[Relationship],
) -> None:
    package_cache = create_package_cache(packages)

    spdx_id_cache = {pkg: spdx_pkg.spdx_id for pkg, spdx_pkg in package_cache.items()}

    document.packages = list(package_cache.values())

    document_relationships = create_document_relationships(document, document.packages)

    process_relationships(document, _relationships, spdx_id_cache, document_relationships)


def format_spdx_sbom(
    *,
    packages: list[Package],
    _relationships: list[Relationship],
    file_format: str,
    config: SbomConfig,
    resolver: Resolver,
) -> None:
    now_utc = datetime.now(UTC)
    namespace, _ = set_namespace_version(config=config, resolver=resolver)
    creation_info = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name=namespace,
        data_license="CC0-1.0",
        document_namespace=get_document_namespace(namespace),
        creators=[Actor(ActorType.TOOL, "Fluid-Labels", None)],
        created=now_utc,
    )

    document = Document(creation_info)

    if not packages and not _relationships:
        add_empty_package(document)
    else:
        add_packages_and_relationships(document, packages, _relationships)

    validation_errors: list[ValidationMessage] = validate_full_spdx_document(document)

    if validation_errors:
        raise SPDXValidationError(validation_errors)

    LOGGER.info(
        "🆗 Valid SPDX %s format, generating output file at %s.%s",
        file_format.upper(),
        config.output,
        file_format,
    )

    write_file(document, f"{config.output}.{file_format}")
    LOGGER.info("✅ Output file successfully generated")



from pydantic import (
    BaseModel,
)

from labels.model.core import (
    Digest,
)


class DpkgFileRecord(BaseModel):
    path: str
    digest: Digest | None = None
    is_config_file: bool | None = None


class DpkgDBEntry(BaseModel):
    package: str
    source: str | None = None
    version: str | None = None
    source_version: str | None = None
    architecture: str | None = None
    maintainer: str | None = None
    installed_size: int | None = None
    description: str | None = None
    provides: list[str] | None = None
    dependencies: list[str] | None = None
    pre_dependencies: list[str] | None = None
    files: list[DpkgFileRecord] | None = None



import logging
from copy import deepcopy
from typing import cast

from labels.artifact.relationship import Relationship, RelationshipType
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import IndexedDict, ParsedValue
from labels.model.core import Package
from labels.pkg.cataloger.generic.parser import Environment
from labels.pkg.cataloger.javascript.package import new_package_lock_v1, new_package_lock_v2

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def get_name_from_path(name: str) -> str:
    return name.split("node_modules/")[-1]


def get_direct_dependencies(package_lock_path: IndexedDict[str, ParsedValue]) -> list[str]:
    all_dependencies: ParsedValue = package_lock_path.get("dependencies", IndexedDict())

    transitives: set[str] = set()
    for details in cast(IndexedDict[str, ParsedValue], all_dependencies).values():
        if not isinstance(details, IndexedDict):
            continue
        reqs = cast(IndexedDict[str, ParsedValue], details.get("requires", EMPTY_DICT))
        for dep in reqs:
            transitives.add(dep)

    return [
        dep
        for dep in cast(IndexedDict[str, ParsedValue], all_dependencies)
        if dep not in transitives
    ]


def get_direct_dependencies_v2_v3(package_lock_path: IndexedDict[str, ParsedValue]) -> list[str]:
    all_dependencies: ParsedValue = package_lock_path.get("packages", IndexedDict())
    if not isinstance(all_dependencies, IndexedDict):
        LOGGER.warning("No direct deps found found in package JSON")
        return []

    result: list[str] = []
    for dep, value in all_dependencies.items():
        if isinstance(value, IndexedDict) and dep == "":
            deps_candidate: ParsedValue = value.get("dependencies", EMPTY_DICT)
            if isinstance(deps_candidate, IndexedDict):
                result.extend(deps_candidate)
    return result


def _solve_sub_dependencies(
    reader: LocationReadCloser,
    sub_deps: IndexedDict[str, ParsedValue],
) -> list[Package]:
    return [
        pkg
        for dep_key, dep_value in sub_deps.items()
        if isinstance(dep_value, IndexedDict)
        and (
            pkg := new_package_lock_v1(
                reader.location,
                dep_key,
                dep_value,
                is_transitive=True,
            )
        )
    ]


def handle_v1(
    reader: LocationReadCloser,
    package_json: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    direct_dependencies = get_direct_dependencies(package_json)
    if not isinstance(deps, IndexedDict):
        LOGGER.warning("No packages found in package JSON")
        return ([], [])
    for dependency_key, dependency_value in deps.items():
        if not isinstance(dependency_value, IndexedDict):
            continue
        name: str = dependency_key
        is_transitive = name not in direct_dependencies
        if pkg := new_package_lock_v1(
            reader.location,
            name,
            dependency_value,
            is_transitive=is_transitive,
        ):
            packages.append(pkg)

        requires = [
            package
            for package in packages
            if package.name
            in cast(IndexedDict[str, ParsedValue], dependency_value.get("requires", EMPTY_DICT))
        ]
        sub_deps: ParsedValue = dependency_value.get("dependencies", EMPTY_DICT)
        if sub_deps and isinstance(sub_deps, IndexedDict):
            packages.extend(_solve_sub_dependencies(reader, sub_deps))
        current_package: Package | None = next(
            (package for package in packages if package.name == dependency_key),
            None,
        )
        if not current_package:
            continue
        relationships.extend(
            [
                Relationship(
                    from_=require,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for require in requires
            ],
        )

    return packages, relationships


def _get_name(dependency_key: str, package_value: IndexedDict[str, ParsedValue]) -> str | None:
    name = dependency_key
    if not name:
        if "name" not in package_value:
            return None
        name = str(package_value["name"])

    # handle alias name
    if "name" in package_value and package_value["name"] != dependency_key:
        name = str(package_value["name"])

    return get_name_from_path(name)


def handle_v2(
    reader: LocationReadCloser,
    package_json: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    pkgs: ParsedValue = package_json.get("packages", IndexedDict())
    if not isinstance(pkgs, IndexedDict):
        LOGGER.warning("No packages found in package JSON")
        return ([], [])
    dependency_map: dict[str, ParsedValue] = {}
    direct_dependencies = get_direct_dependencies_v2_v3(package_json)
    for dependency_key, package_value in pkgs.items():
        if not dependency_key or not isinstance(package_value, IndexedDict):
            continue
        name = _get_name(dependency_key, package_value)
        is_transitive = name not in direct_dependencies
        if pkg := new_package_lock_v2(
            deepcopy(reader.location),
            get_name_from_path(name or dependency_key),
            package_value,
            is_transitive=is_transitive,
        ):
            packages.append(pkg)
            dependencies = package_value.get("dependencies", EMPTY_DICT)
            dependency_map[name or dependency_key] = dependencies

    for pkg in packages:
        dependencies = dependency_map.get(pkg.name, IndexedDict())
        if not isinstance(dependencies, IndexedDict):
            continue
        for dep_name in dependencies:
            dependency_pkg = next((p for p in packages if p.name == dep_name), None)
            if dependency_pkg:
                relationships.append(
                    Relationship(
                        from_=dependency_pkg,
                        to_=pkg,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return packages, relationships


def parse_package_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []

    match package_json.get("lockfileVersion", 1):
        case 1:
            return handle_v1(reader, package_json)
        case 2 | 3:
            return handle_v2(reader, package_json)

    return packages, relationships



import re
from copy import (
    deepcopy,
)
from typing import cast

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)

from .package import (
    package_url,
)

MIX_DEP: re.Pattern[str] = re.compile(
    r"\{:(?P<dep>[\w]*),\s\"~>\s(?P<version>[\d.]+)\".+",
)


def _get_location(
    reader: LocationReadCloser,
    line_number: int,
    *,
    is_dev: bool,
) -> Location:
    location = deepcopy(reader.location)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def parse_mix_exs(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    is_line_deps = False

    for line_number, raw_line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        line = raw_line.strip()
        if line == "defp deps do":
            is_line_deps = True
        elif is_line_deps:
            if line == "end":
                break
            if matched := MIX_DEP.match(line):
                is_dev = ":dev" in line
                pkg_name = cast(str, matched.group("dep"))
                pkg_version = cast(str, matched.group("version"))
                location = _get_location(reader, line_number, is_dev=is_dev)

                packages.append(
                    Package(
                        name=pkg_name,
                        version=pkg_version,
                        type=PackageType.HexPkg,
                        locations=[location],
                        p_url=package_url(pkg_name, pkg_version),
                        metadata=None,
                        language=Language.ELIXIR,
                        licenses=[],
                        is_dev=is_dev,
                    ),
                )
    return packages, relationships



import re
from collections.abc import (
    Iterable,
)
from copy import (
    deepcopy,
)

from gemfileparser import (
    Dependency,
    GemfileParser,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)

from .package import (
    package_url,
)

GEMFILE_DEP: re.Pattern[str] = re.compile(
    r'^\s*(?P<gem>gem ".*?",?( "[><~=]{0,2}\s?[\d\.]+",?){0,2})',
)
NOT_PROD_DEP: re.Pattern[str] = re.compile(
    r":group => \[?[:\w\-, ]*(:development|:test)",
)
NOT_PROD_GROUP: re.Pattern[str] = re.compile(r"(\s*)group :(test|development)")
GEM_LOCK_DEP: re.Pattern[str] = re.compile(
    r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$",
)


def _is_dev_dependency(line: str) -> bool:
    return bool(NOT_PROD_DEP.search(line))


def format_requirements(requirements: list[str]) -> str:
    formatted: str = ""
    if len(requirements) == 0:
        return formatted
    requirements = [req.replace(" ", "") for req in requirements]
    first_req = requirements[0]
    if "~" in first_req:
        if len(requirements) == 1:
            dot_times = first_req.count(".")
            if dot_times <= 1:
                formatted = first_req.replace("~>", "^")
            else:
                formatted = first_req.replace("~>", "~")
        elif len(requirements) == 2:
            sec_req = requirements[1]
            if ">=" in sec_req:
                formatted = sec_req.replace(">=", "^")
            elif "!=" in sec_req:
                sec_ver = sec_req.replace("!=", "")
                first_ver = first_req.replace("!=", "")
                formatted = f">={first_ver} <{sec_ver}  || >{sec_ver}"
    else:
        formatted = " ".join(requirements)
    return formatted


def match_dep_criteria(
    column_list: Iterable[str],
) -> Dependency:
    dep = Dependency()
    for column in column_list:
        for criteria, criteria_regex in GemfileParser.gemfile_regexes.items():
            match = criteria_regex.match(column)
            if match:
                if criteria == "requirement":
                    dep.requirement.append(match.group(criteria))
                else:
                    setattr(dep, criteria, match.group(criteria))
                break
    return dep


def parse_line(in_line: str, *, gem_file: bool) -> tuple[str, str]:
    line: list[str] = []
    if gem_file:
        line = in_line.split(",")
    else:
        line = in_line.split(" ", maxsplit=1)
        if len(line) > 1 and (", " in line[1]):
            line = [line[0], *line[1].split(", ")]

    column_list: list[str] = []
    for column in line:
        stripped_column = (
            column.replace("'", "")
            .replace('"', "")
            .replace("%q<", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
            .strip()
        )
        column_list.append(stripped_column)

    deps = match_dep_criteria(column_list)
    deps_dict = deps.to_dict()
    product: str = deps_dict["name"]
    version: str = format_requirements(deps_dict["requirement"])
    return product, version


def create_dependency(
    line_number: int,
    location: Location,
    matched: re.Match[str],
    *,
    is_dev: bool = False,
) -> Package | None:
    gem_info = GemfileParser.preprocess(matched.group("gem"))[3:]
    product, version = parse_line(gem_info, gem_file=True)
    new_location = deepcopy(location)
    new_location.scope = Scope.DEV if is_dev else Scope.PROD
    if new_location.coordinates:
        new_location.coordinates.line = line_number
        new_location.dependency_type = DependencyType.DIRECT
    if not version:
        return None
    return Package(
        name=product,
        version=version,
        type=PackageType.GemPkg,
        locations=[new_location],
        p_url=package_url(product, version),
        metadata=None,
        language=Language.RUBY,
        licenses=[],
        is_dev=is_dev,
    )


def parse_gemfile(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    in_group_block = False
    end_line = ""

    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if in_group_block:
            if line == end_line:
                in_group_block = False
            elif (
                (matched := GEMFILE_DEP.search(line))
                and matched
                and (
                    dependency := create_dependency(
                        line_number,
                        reader.location,
                        matched,
                        is_dev=_is_dev_dependency(line),
                    )
                )
            ):
                packages.append(dependency)
            continue

        if match_group := NOT_PROD_GROUP.search(line):
            in_group_block = True
            end_line = f"{match_group.group(1)}end"
            continue

        if (
            (matched := GEMFILE_DEP.search(line))
            and matched
            and (
                dependency := create_dependency(
                    line_number,
                    reader.location,
                    matched,
                    is_dev=_is_dev_dependency(line),
                )
            )
        ):
            packages.append(dependency)

    return packages, []



# Magic numbers
HASH_MAGIC_NUMBER = 0x00061561
HASH_MAGIC_NUMBER_BE = 0x61150600

# Sizes in bytes
# The size (in bytes) of an in-page offset
HASH_INDEX_ENTRY_SIZE = 2
# All DB pages have the same sized header (in bytes)
PAGE_HEADER_SIZE = 26
# Hash off-page size (in bytes)
HASH_OFF_PAGE_SIZE = 12

# All page types supported
# Source: https://github.com/berkeleydb/libdb/blob/v5.3.28/src/dbinc
# /db_page.h#L35-L53

# Hash pages created pre 4.6 (DEPRECATED)
HASH_UNSORTED_PAGE_TYPE = 2
OVERFLOW_PAGE_TYPE = 7
HASH_METADATA_PACKAGE_TYPE = 8
# Sorted hash page
HASH_PAGE_TYPE = 13

# Hash off-index page type (aka HOFFPAGE)
# Source: https://github.com/berkeleydb/libdb/blob/v5.3.28/src/dbinc
# /db_page.h#L569-L573
HASH_OF_INDEX_PAGE_TYPE = 3

# PageType is an alias for uint8 in Go.


VALID_PAGE_SIZES = {
    512,
    1024,
    2048,
    4096,
    8192,
    16384,
    32768,
    65536,
}



import logging
from fnmatch import (
    fnmatch,
)
from zipfile import (
    BadZipFile,
    ZipFile,
    ZipInfo,
)

LOGGER = logging.getLogger(__name__)


def normalize_zip_entry_name(entry: str, *, case_insensitive: bool) -> str:
    if case_insensitive:
        entry = entry.lower()
    if not entry.startswith("/"):
        entry = "/" + entry

    return entry


def new_zip_file_manifest(archive_path: str) -> list[ZipInfo]:
    try:
        with ZipFile(archive_path, "r") as myzip:
            return myzip.infolist()
    except BadZipFile:
        return []


def zip_glob_match(
    manifest: list[ZipInfo],
    *,
    case_sensitive: bool,
    patterns: tuple[str, ...],
) -> list[str]:
    result = []

    for pattern in patterns:
        for entry in manifest:
            normalized_entry = normalize_zip_entry_name(
                entry.filename,
                case_insensitive=case_sensitive,
            )
            if entry.filename.endswith(pattern):
                result.append(entry.filename)
            lower_pattern = pattern.lower() if case_sensitive else pattern
            if fnmatch(normalized_entry, lower_pattern):
                result.append(entry.filename)
    return result



import requests
from bs4 import (
    BeautifulSoup,
    NavigableString,
    Tag,
)
from furl import (
    furl,
)

from labels.internal.cache import (
    dual_cache,
)


def format_maven_pom_ulr(
    group_id: str,
    artifact_id: str,
    version: str,
    maven_base_url: str | None = None,
) -> str:
    maven_base_url = maven_base_url or "https://repo1.maven.org/maven2"
    url_object = furl(maven_base_url)
    artifact_pom = f"{artifact_id}-{version}.pom"
    url_object.path.add([*(group_id.split(".")), artifact_id, version, artifact_pom])
    return url_object.url


def get_pom_from_maven_repo(
    *,
    group_id: str,
    artifact_id: str,
    version: str,
    maven_base_url: str | None = None,
) -> BeautifulSoup | None:
    request_url = format_maven_pom_ulr(group_id, artifact_id, version, maven_base_url)
    request = dual_cache(requests.get)(request_url, timeout=30)
    if request.status_code != 200:
        return None

    pom_text = request.text
    return BeautifulSoup(pom_text, features="html.parser")


def get_dependency_version(
    dependency_management: Tag,
    group_id: str,
    artifact_id: str,
) -> str | None:
    """Retrieve specific dependency version from dependency management."""
    for dependency in dependency_management.find_all("dependency"):
        if (
            (dependency_groupid_node := dependency.find_next("groupid"))
            and (dependency_artifactid_node := dependency.find_next("artifactid"))
            and (dependency_version_node := dependency.find_next("version"))
            and dependency_groupid_node.get_text() == group_id
            and dependency_artifactid_node.get_text() == artifact_id
        ):
            return dependency_version_node.get_text()
    return None


def _get_text(parent: Tag | NavigableString, name: str) -> str | None:
    element = parent.find_next(name)
    if element:
        return element.get_text()
    return None


def get_parent_information(
    parent: Tag | NavigableString,
) -> tuple[str | None, str | None, str | None]:
    """Extract parent artifact information."""
    parent_group_id = _get_text(parent, "groupid")
    parent_artifact_id = _get_text(parent, "artifactid")
    parent_version = _get_text(parent, "version")
    return parent_group_id, parent_artifact_id, parent_version


def process_pom(
    parent_pom: Tag,
    group_id: str,
    artifact_id: str,
) -> tuple[str | None, str | None, str | None, str | None]:
    """Process POM to either get the version or update parent details."""
    project = parent_pom.project
    if not project:
        return None, None, None, None

    dependency_management = project.find_next("dependencymanagement")
    if dependency_management and isinstance(dependency_management, Tag):
        version = get_dependency_version(dependency_management, group_id, artifact_id)
        if version:
            return version, None, None, None

    parent = project.find_next("parent")
    if parent:
        return None, *get_parent_information(parent)

    return None, None, None, None


def recursively_find_versions_from_parent_pom(  # noqa: PLR0913
    *,
    group_id: str,
    artifact_id: str,
    parent_group_id: str,
    parent_artifact_id: str,
    parent_version: str,
    maven_base_url: str | None = None,
) -> str | None:
    for _ in range(3):
        parent_pom = get_pom_from_maven_repo(
            group_id=parent_group_id,
            artifact_id=parent_artifact_id,
            version=parent_version,
            maven_base_url=maven_base_url,
        )
        if not parent_pom:
            break

        (
            version,
            new_parent_group_id,
            new_parent_artifact_id,
            new_parent_version,
        ) = process_pom(parent_pom, group_id, artifact_id)
        if version:
            return version
        if new_parent_group_id is None or not new_parent_artifact_id or not new_parent_version:
            break
        parent_group_id, parent_artifact_id, parent_version = (
            new_parent_group_id,
            new_parent_artifact_id,
            new_parent_version,
        )

    return None



from collections.abc import Hashable
from enum import (
    Enum,
)

from pydantic import (
    BaseModel,
    ConfigDict,
)


class RelationshipType(Enum):
    OWNERSHIP_BY_FILE_OVERLAP_RELATIONSHIP = "ownership-by-file-overlap"
    EVIDENT_BY_RELATIONSHIP = "evident-by"
    CONTAINS_RELATIONSHIP = "contains"
    DEPENDENCY_OF_RELATIONSHIP = "dependency-of"
    DESCRIBED_BY_RELATIONSHIP = "described-by"


class Relationship(BaseModel):
    from_: Hashable
    to_: Hashable
    type: RelationshipType
    model_config = ConfigDict(frozen=True)



import os
import re
import uuid
from pathlib import Path
from urllib.parse import (
    urlunparse,
)

from labels.file.resolver import Resolver
from labels.internal.file_resolver.container_image import (
    ContainerImage,
)
from labels.model.core import (
    HealthMetadata,
    Package,
    SbomConfig,
    SourceType,
)
from labels.sources.directory_source import (
    Directory,
)


def is_valid_email(email: str) -> bool:
    email_regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(email_regex, email) is not None


def get_author_info(
    health_metadata: HealthMetadata,
) -> list[tuple[str | None, str | None]]:
    author_info = []

    if health_metadata.authors:
        authors_list = health_metadata.authors.split(", ")
        for author in authors_list:
            name: str | None = None
            email: str | None = None
            email_match = re.search(r"<([^<>]+)>", author)

            if email_match:
                email_candidate: str = email_match.group(1)
                email = email_candidate if is_valid_email(email_candidate) else None
                name = author.replace(email_match.group(0), "").strip() or None
            else:
                name = author.strip() or None

            author_info.append((name, email))

    return author_info


def sanitize_name(text: str) -> str:
    return re.sub(r"[^a-zA-Z0-9.-]", "-", text)


def get_document_namespace(working_dir: str) -> str:
    input_type = "unknown-source-type"

    if Path(working_dir).is_file():
        input_type = "file"
    elif Path(working_dir).is_dir():
        input_type = "dir"

    unique_id = uuid.uuid4()
    identifier = os.path.join(input_type, str(unique_id))
    if working_dir != ".":
        identifier = os.path.join(input_type, f"{working_dir}-{unique_id}")

    return urlunparse(
        ("https", "fluidattacks.com", identifier, "", "", ""),
    )


def merge_packages(packages: list[Package]) -> list[Package]:
    merged_packages: dict[str | Package, Package] = {}

    for package in packages:
        if package.id_:
            if package.id_ in merged_packages:
                merged_packages[package.id_].locations.extend(
                    x for x in package.locations if x not in merged_packages[package.id_].locations
                )
            else:
                merged_packages[package.id_] = package
        else:
            merged_packages[package] = package

    return list(merged_packages.values())


def process_packages(packages: list[Package]) -> list[Package]:
    return merge_packages(packages)


def set_namespace_version(config: SbomConfig, resolver: Resolver) -> tuple[str, str | None]:
    namespace = ""
    version = None
    if config.source_type == SourceType.DIRECTORY and isinstance(resolver, Directory):
        namespace = resolver.root
    if config.source_type in {SourceType.DOCKER, SourceType.ECR} and isinstance(
        resolver,
        ContainerImage,
    ):
        namespace = resolver.context.image_ref
        version = resolver.context.id
    return namespace, version



import logging
import os
from copy import (
    deepcopy,
)
from ctypes import (
    c_void_p,
    cdll,
)

from pydantic import (
    ValidationError,
)
from tree_sitter import (
    Language as TLanguage,
)
from tree_sitter import (
    Node,
    Parser,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.ruby.package import (
    package_url,
)
from labels.utils.env import (
    TREE_SITTER_PARSERS,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def lang_from_so(path: str, name: str) -> TLanguage:
    lib = cdll.LoadLibrary(os.fspath(path))
    language_function = getattr(lib, f"tree_sitter_{name}")
    language_function.restype = c_void_p
    language_ptr = language_function()
    return TLanguage(language_ptr)


def collect_gem_entries(_content: str) -> tuple[list[Node], list[str]]:
    if TREE_SITTER_PARSERS is None:
        LOGGER.warning(
            (
                "Unable to parse gemfile.lock file because the binary file "
                "for the parser is not available in the environment."
            ),
        )
        return [], []
    so_library_path: str = os.path.join(TREE_SITTER_PARSERS, "gemfilelock.so")
    parser_language = lang_from_so(so_library_path, "gemfilelock")
    parser = Parser(parser_language)
    result = parser.parse(_content.encode("utf-8"))

    dependencies_node = next(
        (node for node in result.root_node.children if node.type == "dependencies"),
        None,
    )
    dependencies: list[str] = []
    if dependencies_node and dependencies_node.children:
        dependencies.extend(
            dependency.children[0].parent.named_children[0].text.decode("utf-8")
            for dependency in dependencies_node.children[1:]
            if (
                dependency.children
                and dependency.children[0].parent
                and dependency.children[0].parent.named_children
                and dependency.children[0].parent.named_children[0].text
            )
        )

    gem_section = next(
        (node for node in result.root_node.children if node.type == "gem_section"),
        None,
    )
    if gem_section and (
        specs := next(
            (x for x in gem_section.children[1].children if x.type == "specs"),
            None,
        )
    ):
        return [x.children[0] for x in specs.children[1:]], dependencies

    return [], dependencies


def parse_gemfile_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    gem_entries, dependencies = collect_gem_entries(reader.read_closer.read())

    packages = _process_packages(gem_entries, reader, dependencies)
    relationships = _process_dependencies(gem_entries, packages)

    return packages, relationships


def _process_packages(
    gem_entries: list,
    reader: LocationReadCloser,
    dependencies: list,
) -> list[Package]:
    packages = []

    for gem_entry in gem_entries:
        gem_name = gem_entry.named_children[0].text.decode("utf-8")
        gem_version = gem_entry.named_children[1].text.decode("utf-8")

        if not gem_name or not gem_version:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = gem_entry.start_point[0] + 1
            location.dependency_type = (
                DependencyType.DIRECT if gem_name in dependencies else DependencyType.TRANSITIVE
            )

        try:
            packages.append(
                Package(
                    name=gem_name,
                    version=gem_version,
                    locations=[location],
                    language=Language.RUBY,
                    licenses=[],
                    p_url=package_url(gem_name, gem_version),
                    type=PackageType.GemPkg,
                    metadata=None,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def _process_dependencies(
    gem_entries: list,
    packages: list[Package],
) -> list[Relationship]:
    relationships = []

    for gem_entry in gem_entries:
        gem_entry_name = gem_entry.named_children[0].text.decode("utf-8")
        _package = next(
            (pkg for pkg in packages if pkg.name == gem_entry_name),
            None,
        )

        if not _package or not gem_entry.parent:
            continue

        for dependency_node in (x for x in gem_entry.parent.children if x.type == "dependency"):
            dependency_name = dependency_node.named_children[0].text.decode(
                "utf-8",
            )

            dependency_package = next(
                (pkg for pkg in packages if pkg.name == dependency_name),
                None,
            )

            if dependency_package:
                relationships.append(
                    Relationship(
                        from_=_package,
                        to_=dependency_package,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships



from pydantic import (
    BaseModel,
)

from labels.file.coordinates import (
    Coordinates,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.scope import (
    Scope,
)


class LocationMetadata(BaseModel):
    annotations: dict[str, str]

    def merge(self, other: "LocationMetadata") -> "LocationMetadata":
        return LocationMetadata(annotations={**self.annotations, **other.annotations})


class LocationData(BaseModel):
    coordinates: Coordinates
    access_path: str

    def __hash__(self) -> int:
        return hash(self.access_path) + hash(self.coordinates.file_system_id)


class Location(BaseModel):
    scope: Scope = Scope.PROD
    coordinates: Coordinates | None = None
    access_path: str | None = None
    annotations: dict[str, str] | None = None
    dependency_type: DependencyType = DependencyType.UNKNOWN

    def with_annotation(self, key: str, value: str) -> "Location":
        if not self.annotations:
            self.annotations = {}
        self.annotations[key] = value
        return self

    def path(self) -> str:
        path = self.access_path or (self.coordinates.real_path if self.coordinates else "") or ""
        return path.strip().replace(" ", "_")


def new_location_from_image(
    access_path: str | None,
    layer_id: str,
    real_path: str | None = None,
) -> Location:
    if access_path and not access_path.startswith("/"):
        access_path = f"/{access_path}"
    return Location(
        coordinates=Coordinates(real_path=real_path or "", file_system_id=layer_id),
        access_path=access_path,
        annotations={},
    )


def new_location(real_path: str) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
        ),
        access_path=real_path,
        annotations={},
    )



import base64
import logging
from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)

from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.advisories import (
    roots as advisories,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.scope import (
    Scope,
)
from labels.internal.collection.types import (
    IndexedDict,
    ParsedValue,
)
from labels.internal.package_information.javascript import (
    NPMPackage,
    NPMPackageLicense,
    get_npm_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.pkg.cataloger.javascript.model import (
    NpmPackageLockEntry,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def handle_licenses(
    licenses: str | list[str | dict[str, str]] | NPMPackageLicense,
) -> list[str]:
    if isinstance(licenses, dict):
        return [licenses["type"]] if "type" in licenses else []
    if isinstance(licenses, list):
        licenses_list = []
        for license_item in licenses:
            if isinstance(license_item, str):
                licenses_list.append(license_item)
            if isinstance(license_item, dict) and license_item["type"]:
                licenses_list.append(license_item["type"])
        return licenses_list
    return [licenses]


def new_package_lock_v1(
    location: Location,
    name: str,
    value: IndexedDict[str, ParsedValue],
    *,
    is_transitive: bool,
) -> Package | None:
    version: str = str(value.get("version", ""))
    if not name or not version:
        return None

    alias_prefix_package_lock = "npm:"
    if version.startswith(alias_prefix_package_lock):
        name, version = version.removeprefix(alias_prefix_package_lock).split(
            "@",
        )
    current_location = deepcopy(location)
    is_dev = value.get("dev", False)
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=str(value.get("resolved")) if value.get("resolved") is not None else None,
                integrity=str(value.get("integrity"))
                if value.get("integrity") is not None
                else None,
                is_dev=bool(value.get("dev", False)),
            )
            if value.get("resolved") and "integrity" in value
            else None,
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def new_package_lock_v2(
    location: Location,
    name: str,
    value: IndexedDict[str, ParsedValue],
    *,
    is_transitive: bool,
) -> Package | None:
    version: str = str(value.get("version", ""))

    if not name or not version:
        return None

    current_location = location
    is_dev = bool(value.get("dev", False))
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=str(value.get("resolved")) if value.get("resolved") is not None else None,
                integrity=str(value.get("integrity"))
                if value.get("integrity") is not None
                else None,
                is_dev=is_dev,
            ),
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def package_url(name: str, version: str) -> str:
    namespace = ""
    fields = name.split("/", 2)
    if len(fields) > 1:
        namespace = fields[0]
        name = fields[1]

    if not name:
        return ""

    return PackageURL(  # type: ignore
        type="npm",
        namespace=namespace,
        name=name,
        version=version,
        qualifiers={},
        subpath="",
    ).to_string()


def _get_author(npm_package: NPMPackage) -> str | None:
    author: str | None = None
    if "author" in npm_package:
        package_author = npm_package["author"]
        if isinstance(package_author, dict) and "name" in package_author:
            author = package_author["name"]
            if "email" in package_author:
                author = f'{author} <{package_author["email"]}>'
        elif package_author and isinstance(package_author, str):
            author = str(package_author)
        return author
    return None


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("npm", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def _get_latest_version_info(
    npm_package: NPMPackage,
    current_version: str,
) -> tuple[str | None, str | None]:
    using_pre_release = "-" in current_version
    latest_version = None
    latest_version_created_at = None

    if using_pre_release:
        latest_version = list(npm_package["versions"].keys())[-1]
        latest_version_created_at = npm_package["time"]["modified"]
    elif npm_package.get("dist-tags"):
        latest_version = npm_package["dist-tags"]["latest"]
        latest_version_created_at = npm_package["time"][latest_version]

    return latest_version, latest_version_created_at


def _get_artifact_info(
    npm_package: NPMPackage,
    current_version: str,
) -> Artifact | None:
    current_package = npm_package["versions"].get(current_version)
    artifact = None

    if current_package:
        with suppress(KeyError):
            digest_value = current_package.get("dist", {}).get("integrity") or None

            if digest_value:
                algorithm, digest_hash = digest_value.split("-", 1)
                if algorithm == "sha512":
                    binary_hash = base64.b64decode(digest_hash)
                    digest_hash = binary_hash.hex()

                artifact = Artifact(
                    url=current_package["dist"]["tarball"],
                    integrity=Digest(
                        algorithm=infer_algorithm(digest_hash),
                        value=digest_hash,
                    ),
                )

    return artifact


def _set_health_metadata(package: Package, npm_package: NPMPackage) -> None:
    latest_version, latest_version_created_at = _get_latest_version_info(
        npm_package,
        package.version,
    )
    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        latest_version_created_at=latest_version_created_at,
        artifact=_get_artifact_info(npm_package, package.version),
        authors=_get_author(npm_package),
    )


def complete_package(package: Package) -> Package:
    _update_advisories(package)

    npm_package = get_npm_package(package.name)
    if not npm_package:
        return package

    _set_health_metadata(package, npm_package)

    licenses = npm_package.get("license")
    if licenses and isinstance(licenses, (str, list, dict)):  # noqa: UP038
        package.licenses = validate_licenses(handle_licenses(licenses))

    return package



from typing import (
    TypedDict,
)

from labels.internal.package_information.api_interface import (
    make_get,
)


class OriginInfo(TypedDict):
    VCS: str
    URL: str
    Ref: str
    Hash: str


class VersionInfo(TypedDict):
    Version: str
    Time: str
    Origin: OriginInfo | None


class LicenseLinks(TypedDict):
    self: str
    git: str
    html: str


class LicenseInfo(TypedDict):
    key: str
    name: str
    spdx_id: str
    url: str | None
    node_id: str


class LicenseData(TypedDict):
    name: str
    path: str
    sha: str
    size: int
    url: str
    html_url: str
    git_url: str
    download_url: str
    type: str
    content: str
    encoding: str
    _links: LicenseLinks
    license: LicenseInfo


def fetch_license_info(repo_path: str) -> LicenseData | None:
    base_url = "https://api.github.com/repos"
    url = f"{base_url}/{repo_path}/license"

    response: LicenseData | None = make_get(url)
    return response


def fetch_latest_version_info(module_path: str) -> VersionInfo | None:
    base_url = "https://proxy.golang.org"
    url = f"{base_url}/{module_path.lower()}/@latest"

    response: VersionInfo | None = make_get(url)
    return response



import logging
import os
import stat
import tempfile
import zipfile
from pathlib import (
    Path,
)

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomProperties,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def is_safe_path(base_path: str, target_path: str) -> bool:
    base_path = os.path.normpath(base_path)
    target_path = os.path.normpath(target_path)
    return os.path.commonpath([base_path]) == os.path.commonpath([base_path, target_path])


def safe_extract(apk_file: zipfile.ZipFile, destination: str) -> None:
    for file_info in apk_file.infolist():
        file_name = file_info.filename
        if Path(file_name).is_absolute() or file_name.startswith(("..", "./")):
            continue

        target_path = os.path.join(destination, file_name)

        if not is_safe_path(destination, target_path):
            continue

        if (file_info.external_attr >> 16) & stat.S_IFLNK:
            continue

        try:
            apk_file.extract(file_name, destination)
        except Exception:
            LOGGER.exception("Error extracting %s", file_name)


def parse_apk(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    with tempfile.TemporaryDirectory() as output_folder:
        try:
            with zipfile.ZipFile(reader.read_closer.name, "r") as apk_file:
                safe_extract(apk_file, output_folder)
        except zipfile.BadZipFile:
            return packages, []
        files: list[str] = []
        meta_dir = os.path.join(output_folder, "META-INF")
        if Path(meta_dir).exists():
            files = [
                os.path.join(meta_dir, file)
                for file in os.listdir(meta_dir)
                if file.endswith(".version")
            ]
        for file in files:
            file_path = Path(file)
            with file_path.open(encoding="utf-8") as version_reader:
                version = version_reader.read().strip()
            parts = file_path.name.replace(".version", "").split("_", 1)
            group_id = parts[0]
            artifact_id = parts[1]

            if any(not value for value in (artifact_id, version, group_id)):
                continue

            java_archive = JavaArchive(
                pom_properties=JavaPomProperties(
                    group_id=group_id,
                    artifact_id=artifact_id,
                    version=version,
                ),
            )

            try:
                packages.append(
                    Package(
                        name=artifact_id,
                        version=version,
                        licenses=[],
                        locations=[reader.location],
                        language=Language.JAVA,
                        type=PackageType.JavaPkg,
                        metadata=java_archive,
                        p_url=package_url(artifact_id, version, java_archive),
                    ),
                )
            except ValidationError as ex:
                LOGGER.warning(
                    "Malformed package. Required fields are missing or data "
                    "types are incorrect.",
                    extra={
                        "extra": {
                            "exception": format_exception(str(ex)),
                            "location": reader.location.path(),
                        },
                    },
                )
                continue

    return packages, []



from os import (
    environ,
)
from typing import (
    Literal,
)


def _get_artifact(env_var: str) -> str | None:
    if value := environ.get(env_var):
        return value
    return None


TREE_SITTER_PARSERS = _get_artifact("TREE_SITTER_PARSERS_DIR")


def guess_environment() -> Literal["development", "production"]:
    return (
        "production"
        if environ.get("AWS_BATCH_JOB_ID") or environ.get("CI_COMMIT_REF_NAME") == "trunk"
        else "development"
    )



import logging
import re
from copy import (
    deepcopy,
)
from typing import (
    NotRequired,
    TypedDict,
)

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.javascript.model import (
    YarnLockEntry,
)
from labels.pkg.cataloger.javascript.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _resolve_pair(line: str) -> tuple[str, str]:
    line = line.strip()
    if ": " in line:
        key, value = line.split(": ")
        return key.strip(), value.strip()

    key, value = line.split(" ", maxsplit=1)
    return key.strip(), value.strip()


def _count_indentation(line: str) -> int:
    # Stripping the leading spaces and comparing the length difference
    return len(line) - len(line.lstrip(" "))


def _is_start_of_list(
    current_package: str | None,
    current_package_version: str | None,
    line: str,
) -> bool:
    return bool(
        current_package and current_package_version and ":" in line and line.strip().endswith(":"),
    )


def _is_list_item(
    current_package: str | None,
    current_package_version: str | None,
    current_key: str | None,
    current_indentation: int | None,
    line: str,
) -> bool:
    return bool(
        current_package
        and current_package_version
        and current_key
        and current_indentation
        and _count_indentation(line) > current_indentation,
    )


class YarnPackage(TypedDict):
    checksum: str
    dependencies: NotRequired[list[tuple[str, str]]]
    integrity: NotRequired[str]
    line: int
    resolution: NotRequired[str]
    resolved: NotRequired[str]
    version: str


def parse_current_package(line: str, index: int) -> tuple[str | None, int | None]:
    line = line.strip()
    if match_ := re.match(r'^"?((?:@\w[\w\-\.]*/)?\w[\w\-\.]*)@', line):
        current_package = match_.groups()[0]
        current_package_line = index
    else:
        current_package = None
        current_package_line = None

    return current_package, current_package_line


def _parse_yarn_file(yarn_lock_content: str) -> dict[tuple[str, str], YarnPackage]:
    yarn_lock_lines = yarn_lock_content.strip().split("\n")

    # Dictionary to store the parsed yarn lock data
    parsed_yarn_lock = {}

    # Temporary variables for current package and dependencies
    current_package: str | None = None
    current_package_version: str | None = None
    current_indentation = None
    current_key = None

    # Iterate through each line and parse the content
    for index, line in enumerate(yarn_lock_lines, 1):
        if not line:
            current_indentation = None
            continue
        if line.startswith("#"):
            continue
        if not line.startswith(" "):
            current_package, current_package_line = parse_current_package(line, index)
            current_package_version = None
        if current_package and line.strip().startswith("version"):
            _, raw_version = _resolve_pair(line)
            current_package_version = raw_version.strip('"')
            parsed_yarn_lock[(current_package, current_package_version)] = {
                "line": current_package_line,
                "version": current_package_version,
            }
        elif _is_start_of_list(current_package, current_package_version, line):
            current_indentation = _count_indentation(line)
            current_key = line.strip().split(":")[0]
            parsed_yarn_lock[(current_package, current_package_version)][  # type: ignore
                current_key
            ] = []  # type: ignore
        elif _is_list_item(
            current_package,
            current_package_version,
            current_key,
            current_indentation,
            line,
        ):
            parsed_yarn_lock[(current_package, current_package_version)][  # type: ignore
                current_key  # type: ignore
            ].append(_resolve_pair(line))
        elif current_package and current_package_version:
            current_indentation = None
            key, value = _resolve_pair(line)
            parsed_yarn_lock[(current_package, current_package_version)][key] = value.strip('"')

    return parsed_yarn_lock  # type: ignore


def _get_name(pkg_info: tuple[str, str], item: YarnPackage) -> str:
    if resolution := item.get("resolution"):
        is_scoped_package = resolution.startswith("@")
        if is_scoped_package:
            return f"@{resolution.split('@')[1]}"
        return resolution.split("@")[0]

    return pkg_info[0]


def _extract_packages(
    parsed_yarn_lock: dict[tuple[str, str], YarnPackage],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    for pkg_info, item in parsed_yarn_lock.items():
        name = _get_name(pkg_info, item)
        version = item.get("version")

        if not name or not version:
            continue

        current_location = deepcopy(reader.location)
        if current_location.coordinates:
            current_location.coordinates.line = item["line"]
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[current_location],
                    language=Language.JAVASCRIPT,
                    licenses=[],
                    type=PackageType.NpmPkg,
                    p_url=package_url(name, version),
                    metadata=YarnLockEntry(
                        resolved=item.get("resolved"),
                        integrity=item.get("integrity"),
                    ),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
            continue

    return packages


def _extract_relationships(
    parsed_yarn_lock: dict[tuple[str, str], YarnPackage],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    for pkg_info, item in parsed_yarn_lock.items():
        current_pkg = next(
            (package for package in packages if package.name == _get_name(pkg_info, item)),
            None,
        )

        if current_pkg is None:
            continue

        if "dependencies" in item:
            for raw_dep_name, _ in item["dependencies"]:
                dep_name = raw_dep_name.strip('"')
                # TO-DO: check if the version matches
                if dep := next(
                    (package for package in packages if package.name == dep_name),
                    None,
                ):
                    relationships.append(
                        Relationship(
                            from_=dep,
                            to_=current_pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )
    return relationships


def parse_yarn_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    parsed_yarn_lock = _parse_yarn_file(reader.read_closer.read())
    packages = _extract_packages(parsed_yarn_lock, reader)
    relationships = _extract_relationships(parsed_yarn_lock, packages)
    return packages, relationships



import logging
from contextlib import suppress
from datetime import datetime
from typing import cast

from bs4 import BeautifulSoup, Tag
from packageurl import PackageURL
from pydantic import BaseModel, ValidationError

from labels.advisories import images as advisories
from labels.file.location import Location
from labels.internal.package_information.alpine import get_package_versions_html
from labels.linux.release import Release
from labels.model.core import Digest, HealthMetadata, Language, Package, PackageType
from labels.utils.licenses.validation import validate_licenses
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class ApkFileRecord(BaseModel):
    path: str
    owner_uid: str | None = None
    owner_gid: str | None = None
    permissions: str | None = None
    digest: Digest | None = None


class ApkDBEntry(BaseModel):
    package: str
    origin_package: str | None
    maintainer: str | None
    version: str
    architecture: str | None
    url: str | None
    description: str | None
    size: str
    installed_size: str | None
    dependencies: list[str]
    provides: list[str]
    checksum: str | None
    git_commit: str | None
    files: list[ApkFileRecord]


class ParsedData(BaseModel):
    apk_db_entry: ApkDBEntry
    license: str | None


def package_url(entry: ApkDBEntry, distro: Release | None) -> str:
    qualifiers = {"arch": entry.architecture or ""} if entry else {}

    if entry and entry.origin_package != entry.package and entry.origin_package:
        qualifiers["upstream"] = entry.origin_package
    distro_qualifiers = []

    if distro and distro.id_:
        qualifiers["distro_id"] = distro.id_
        distro_qualifiers.append(distro.id_)

    if distro and distro.version_id:
        qualifiers["distro_version_id"] = distro.version_id
        distro_qualifiers.append(distro.version_id)
    elif distro and distro.build_id:
        distro_qualifiers.append(distro.build_id)

    if distro_qualifiers:
        qualifiers["distro"] = "-".join(distro_qualifiers)

    return PackageURL(
        type="apk",
        namespace=distro.id_.lower() if distro and distro.id_ else "",
        name=entry.package,
        version=entry.version,
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def new_package(
    data: ParsedData,
    release: Release | None,
    db_location: Location,
) -> Package | None:
    name = data.apk_db_entry.package
    version = data.apk_db_entry.version

    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            locations=[db_location],
            licenses=validate_licenses(data.license.split(" ")) if data.license else [],
            p_url=package_url(data.apk_db_entry, release),
            type=PackageType.ApkPkg,
            metadata=data.apk_db_entry,
            found_by=None,
            health_metadata=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": db_location.path(),
                },
            },
        )
        return None


def _get_latest_version_and_latest_version_created_at(
    package: Package,
    distro_version: str | None,
    arch: str | None,
) -> tuple[str, datetime | None] | tuple[None, None]:
    html_content = get_package_versions_html(package.name, distro_version, arch)

    if not html_content:
        return None, None

    parsed_content = BeautifulSoup(html_content, features="html.parser")
    version_items: list[Tag] = list(parsed_content.find_all("td", {"class": "version"}))

    if version_items:
        latest_version = version_items[0].text.strip()
        latest_version_created_at = None
        with suppress(IndexError):
            parent_tr = next(iter(version_items[0].fetchPrevious("tr", limit=1)))
            if build_date_tag := parent_tr.find_next("td", {"class": "bdate"}):
                latest_version_created_at = datetime.fromisoformat(build_date_tag.text.strip())
        return latest_version, latest_version_created_at

    return None, None


def _set_health_metadata(package: Package, arch: str | None, distro_version: str | None) -> None:
    authors = (
        package.metadata if package.metadata and hasattr(package.metadata, "maintainer") else None
    )
    (
        latest_version,
        latest_version_created_at,
    ) = _get_latest_version_and_latest_version_created_at(package, distro_version, arch)

    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        latest_version_created_at=latest_version_created_at,
        authors=cast(str, authors.maintainer),
    )


def complete_package(
    package: Package,
    distro_id: str | None = None,
    distro_version: str | None = None,
    arch: str | None = None,
) -> Package:
    pkg_advisories = advisories.get_package_advisories(
        package,
        distro_id=distro_id,
        distro_version=distro_version,
    )

    if pkg_advisories:
        package.advisories = pkg_advisories

    _set_health_metadata(package, arch, distro_version)

    return package



import json
import logging
import os
import re
import shutil
import subprocess  # nosec
import tarfile
import tempfile
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict

from labels.utils.file import extract_tar_file

LOGGER = logging.getLogger(__name__)


class LayerData(BaseModel):
    mimetype: str
    digest: str
    size: int
    annotations: dict[str, str] | None
    model_config = ConfigDict(frozen=True)


class ImageMetadata(BaseModel):
    name: str
    digest: str
    repotags: list[str]
    created: str
    dockerversion: str
    labels: dict[str, str] | None
    architecture: str
    os: str
    layers: list[str]
    layersdata: list[LayerData]
    env: list[str]
    image_ref: str = ""
    model_config = ConfigDict(frozen=True)


class LayerInfo(BaseModel):
    id_: str
    tar_full_path: str
    metadata: dict[str, Any]


class ImageContext(BaseModel):
    id: str
    name: str
    publisher: str | None
    arch: str
    size: str
    full_extraction_dir: str
    layers_dir: str
    manifest: dict[str, Any]
    image_ref: str

    def get_layer_info(self, layer_id: str) -> LayerInfo | None:
        if layer_metadata := next(x for x in self.manifest["layers"] if x["id"] == layer_id):
            return LayerInfo(
                id_=layer_id,
                tar_full_path=os.path.join(self.layers_dir, layer_metadata),
                metadata=layer_metadata,
            )
        return None


def custom_object_hook(json_object: dict[str, Any]) -> ImageMetadata | dict:
    if "Name" in json_object and "Digest" in json_object and "RepoTags" in json_object:
        layersdata = [
            LayerData(
                mimetype=layer_data["MIMEType"],
                digest=layer_data["Digest"],
                size=layer_data["Size"],
                annotations=layer_data.get("Annotations"),
            )
            for layer_data in json_object["LayersData"]
        ]
        return ImageMetadata(
            name=json_object["Name"],
            digest=json_object["Digest"],
            repotags=json_object["RepoTags"],
            created=json_object["Created"],
            dockerversion=json_object["DockerVersion"],
            labels=json_object["Labels"],
            architecture=json_object["Architecture"],
            os=json_object["Os"],
            layers=json_object["Layers"],
            layersdata=layersdata,
            env=json_object["Env"],
        )
    return json_object


def _extract_present_layers(
    layers_dir: str,
    output_dir: str,
) -> None:
    for file_path in sorted(os.listdir(layers_dir)):
        layer_tar_path = os.path.join(layers_dir, file_path, "layer.tar")
        if (
            Path(layer_tar_path).exists()
            and Path(layer_tar_path).is_file()
            and tarfile.is_tarfile(layer_tar_path)
        ):
            Path(os.path.join(output_dir, file_path)).mkdir(
                parents=True,
                exist_ok=True,
            )
            extract_tar_file(
                layer_tar_path,
                os.path.join(output_dir, file_path),
            )


def copy_image(  # noqa: PLR0913
    image_ref: str,
    dest_path: str,
    *,
    username: str | None = None,
    password: str | None = None,
    aws_creds: str | None = None,
    token: str | None = None,
) -> bool:
    skopeo_path = shutil.which("skopeo")
    formated_image_ref = format_image_ref(image_ref)
    if not skopeo_path or not formated_image_ref:
        return False
    command_args = [
        skopeo_path,
        "copy",
        "--dest-decompress",
        "--src-tls-verify=false",
        "--insecure-policy",
        "--override-os",
        "linux",
        formated_image_ref,
        f"dir:{dest_path}",
    ]
    if username and password:
        command_args.extend(
            ["--src-username", username, "--src-password", password],
        )
    elif token:
        command_args.extend(["--src-registry-token", token])
    elif aws_creds:
        command_args.append(f"--src-creds={aws_creds}")

    with subprocess.Popen(  # noqa: S603
        command_args,
        shell=False,
        stdout=subprocess.PIPE,
    ) as proc:
        exit_code = proc.wait()
        return exit_code == 0


def extract_docker_image(  # noqa: PLR0913
    image: ImageMetadata,
    output_dir: str,
    *,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> tuple[str, dict[str, Any]]:
    layers_dir_temp = tempfile.mkdtemp()
    copy_image(
        format_image_ref(image.image_ref, daemon=daemon) or image.image_ref,
        layers_dir_temp,
        username=username,
        password=password,
        token=token,
        aws_creds=aws_creds,
    )
    with Path(os.path.join(layers_dir_temp, "manifest.json")).open(
        encoding="utf-8",
    ) as json_reader:
        manifest = json.load(json_reader)
    with Path(
        os.path.join(
            layers_dir_temp,
            manifest["config"]["digest"].replace("sha256:", ""),
        ),
    ).open(
        encoding="utf-8",
    ) as json_reader:
        manifest["config_full"] = json.load(json_reader)

    for layer in manifest["layers"]:
        layer_digest = layer["digest"].replace("sha256:", "")
        layer_tar_path = os.path.join(
            layers_dir_temp,
            layer_digest,
        )
        if Path(layer_tar_path).exists():
            Path(os.path.join(output_dir, layer["digest"])).mkdir(
                parents=True,
                exist_ok=True,
            )
            extract_tar_file(
                layer_tar_path,
                os.path.join(output_dir, layer["digest"]),
            )

    return layers_dir_temp, manifest


def format_image_ref(image_ref: str, *, daemon: bool = False) -> str | None:
    image_ref_pattern = (
        r"^(?:(?P<host>[\w\.\-]+(?:\:\d+)?)/)?"
        r"(?P<namespace>(?:[\w\.\-]+(?:/[\w\.\-]+)*)?/)?"
        r"(?P<image>[\w\.\-]+)(?::(?P<tag>[\w\.\-]+))?(?:@"
        r"(?P<digest>sha256:[A-Fa-f0-9]{64}))?$"
    )
    prefix_to_use = "docker-daemon:" if daemon else "docker://"
    prefix_used: str | None = None
    prefixes = ["docker://", "docker-daemon:"]
    for prefix in prefixes:
        if image_ref.startswith(prefix):
            image_ref = image_ref.replace(prefix, "", 1)
            prefix_used = prefix
            break

    prefix_to_use = prefix_used or prefix_to_use

    if re.match(image_ref_pattern, image_ref):
        return f"{prefix_to_use}{image_ref}"

    LOGGER.error("Invalid image reference: %s", image_ref_pattern)
    return None


def get_docker_image(  # noqa: PLR0913
    image_ref: str,
    *,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageMetadata | None:
    skopeo_path = shutil.which("skopeo")
    formated_image_ref = format_image_ref(image_ref, daemon=daemon)
    if not skopeo_path or not formated_image_ref:
        return None
    command_args = [
        skopeo_path,
        "inspect",
        "--tls-verify=false",
        "--override-os",
        "linux",
        formated_image_ref,
    ]

    if username and password:
        command_args.extend(["--username", username, "--password", password])

    elif token:
        command_args.append(f"--registry-token={token}")

    elif aws_creds:
        command_args.append(f"--creds={aws_creds}")

    try:
        result = subprocess.run(  # noqa: S603
            command_args,
            check=True,
            capture_output=True,
            text=True,
        )

        image_metadata: ImageMetadata = json.loads(
            result.stdout,
            object_hook=custom_object_hook,
        )
        if image_metadata:
            image_metadata = image_metadata.model_copy(
                update={"image_ref": image_ref},
            )

        return image_metadata
    except subprocess.CalledProcessError as error:
        error_message = f"An error occurred: {error.stderr.strip()}"
        LOGGER.exception(error_message)
        return None


def get_image_context(  # noqa: PLR0913
    *,
    image: ImageMetadata,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageContext | None:
    temp_dir = tempfile.mkdtemp()
    layers_dir, manifest = extract_docker_image(
        image,
        temp_dir,
        username=username,
        password=password,
        token=token,
        aws_creds=aws_creds,
        daemon=daemon,
    )

    return ImageContext(
        id=image.digest,
        name=image.name,
        publisher="",
        arch=image.architecture,
        size=str(sum(x.size for x in image.layersdata)),
        full_extraction_dir=temp_dir,
        layers_dir=layers_dir,
        manifest=manifest,
        image_ref=image.image_ref,
    )



import logging
import zipfile
from collections.abc import Callable
from contextlib import (
    suppress,
)


def traverse_files_in_zip(
    archive_path: str,
    visitor: Callable[[zipfile.ZipInfo], None],
    *paths: str,
) -> None:
    """Traverse files in a zip file applying a visitor function to each file."""
    with zipfile.ZipFile(archive_path, "r") as zip_reader:
        for path in paths:
            try:
                visitor(zip_reader.getinfo(path))
            except KeyError:
                logging.exception("Unable to find file: %s", path)


def contents_from_zip(archive_path: str, *paths: str) -> dict[str, str]:
    """Extract specified files from a zip archive and return their contents."""
    results: dict[str, str] = {}

    if not paths:
        return results  # Return empty result if no paths are specified

    def visitor(file: zipfile.ZipInfo) -> None:
        """Visitor function to read the contents of a file in the zip."""
        if file.is_dir():
            logging.error("Unable to extract directories, only files: %s", file.filename)
            return
        with zipfile.ZipFile(archive_path, "r") as zip_reader, zip_reader.open(file) as file_data:
            content = file_data.read()  # Read the content of the file
            with suppress(UnicodeDecodeError):
                results[file.filename] = content.decode(
                    "utf-8",
                )  # Assuming UTF-8 encoding for simplicity

    traverse_files_in_zip(archive_path, visitor, *paths)
    return results



from labels.advisories import roots as advisories
from labels.internal.package_information.dart import (
    PubPackage,
    PubPackageVersion,
    get_pub_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Package,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)


def _get_current_package(pub_package: PubPackage, version: str) -> PubPackageVersion | None:
    return next((v for v in pub_package["versions"] if v["version"] == version), None)


def _get_authors(pub_package: PubPackage) -> str | None:
    return next(
        (
            version["pubspec"]["author"]
            for version in reversed(pub_package["versions"])
            if "author" in version["pubspec"]
        ),
        None,
    )


def _get_artifact(current_package: PubPackageVersion) -> Artifact:
    digest_value = current_package["archive_sha256"]
    return Artifact(
        url=current_package["archive_url"],
        integrity=Digest(
            value=digest_value,
            algorithm=infer_algorithm(digest_value),
        ),
    )


def _set_health_metadata(
    package: Package,
    pub_package: PubPackage,
    current_package: PubPackageVersion | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=pub_package["latest"]["version"],
        latest_version_created_at=pub_package["latest"]["published"],
        authors=_get_authors(pub_package),
        artifact=_get_artifact(current_package) if current_package else None,
    )


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("pub", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    pub_package = get_pub_package(package.name)
    if not pub_package:
        return package

    current_package = _get_current_package(pub_package, package.version)

    _set_health_metadata(package, pub_package, current_package)

    return package



from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from collections.abc import ItemsView

from labels.artifact.relationship import Relationship
from labels.file.dependency_type import DependencyType
from labels.file.location import Location
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.file.scope import Scope
from labels.internal.collection.json import parse_json_with_tree_sitter
from labels.internal.collection.types import IndexedDict, ParsedValue
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment
from labels.pkg.cataloger.php.package import package_url

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool,
) -> list[Package]:
    if not dependencies:
        return []

    general_location = _get_location(
        reader.location,
        dependencies.position.start.line,
    )
    general_location.scope = Scope.DEV if is_dev else Scope.PROD
    items: ItemsView[str, ParsedValue] = dependencies.items()
    return [
        Package(
            name=name,
            version=version,
            locations=[general_location],
            language=Language.PHP,
            licenses=[],
            type=PackageType.PhpComposerPkg,
            p_url=package_url(name, version),
            is_dev=is_dev,
        )
        for name, version in items
        if isinstance(version, str)
    ]


def parse_composer_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        content.get("require", EMPTY_DICT),
    )
    dev_deps: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        content.get("require-dev", EMPTY_DICT),
    )
    packages = [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



from enum import (
    Enum,
)

from labels.linux.release import (
    Release,
)

PURL_QUALIFIER_DISTRO = "distro"


def purl_qualifiers(
    qualifiers: dict[str, str | None],
    release: Release | None = None,
) -> dict[str, str]:
    # Handling distro qualifiers
    if release:
        distro_qualifiers = []
        if release.id_:
            distro_qualifiers.append(release.id_)
        if release.version_id:
            distro_qualifiers.append(release.version_id)
        elif release.build_id:
            distro_qualifiers.append(release.build_id)

        if distro_qualifiers:
            qualifiers[PURL_QUALIFIER_DISTRO] = "-".join(distro_qualifiers)

    return {
        key: qualifiers.get(key, "") or ""
        for key in sorted(qualifiers.keys())
        if qualifiers.get(key)
    }


class Algorithm(Enum):
    SHA1 = "sha1"
    SHA224 = "sha224"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"
    SHA3_256 = "sha3-256"
    SHA3_384 = "sha3-384"
    SHA3_512 = "sha3-512"
    BLAKE2B_256 = "blake2b-256"
    BLAKE2B_384 = "blake2b-384"
    BLAKE2B_512 = "blake2b-512"
    BLAKE3 = "blake3"
    MD2 = "md2"
    MD4 = "md4"
    MD5 = "md5"
    ADLER32 = "adler32"
    MD6 = "md6"


algorithm_length: dict[Algorithm, str] = {
    Algorithm.SHA1: "40",
    Algorithm.SHA224: "56",
    Algorithm.SHA256: "64",
    Algorithm.SHA384: "96",
    Algorithm.SHA512: "128",
    Algorithm.SHA3_256: "64",
    Algorithm.SHA3_384: "96",
    Algorithm.SHA3_512: "128",
    Algorithm.BLAKE2B_256: "64",
    Algorithm.BLAKE2B_384: "96",
    Algorithm.BLAKE2B_512: "128",
    Algorithm.BLAKE3: "256",
    Algorithm.MD2: "32",
    Algorithm.MD4: "32",
    Algorithm.MD5: "32",
    Algorithm.MD6: "512",
    Algorithm.ADLER32: "8",
}


def infer_algorithm(digest_value: str | None) -> str | None:
    if digest_value:
        for algorithm, length in algorithm_length.items():
            if len(digest_value) == int(length):
                return algorithm.value
    return None



from typing import TypedDict, cast

import requests

from labels.internal.cache import dual_cache


class NugetCatalogEntry(TypedDict, total=False):
    authors: str
    created: str
    description: str
    id: str
    isPrerelease: bool
    lastEdited: str
    licenseExpression: str
    listed: bool
    packageHash: str
    packageHashAlgorithm: str
    packageSize: int
    projectUrl: str
    published: str
    requireLicenseAcceptance: bool
    title: str
    verbatimVersion: str
    version: str
    tags: list[str]


@dual_cache
def get_nuget_package(package_name: str, version: str | None = None) -> NugetCatalogEntry | None:
    package_name = package_name.lower()
    base_url = f"https://api.nuget.org/v3/registration5-gz-semver2/{package_name}"
    if version:
        base_url += f"/{version}"
    else:
        base_url += "/index"
    base_url += ".json"
    response = requests.get(base_url, timeout=30)
    if response.status_code != 200:
        return None

    package_data = response.json()

    if version:
        catalog_url = package_data.get("catalogEntry")
        if not isinstance(catalog_url, str):
            return None
        package_data = requests.get(catalog_url, timeout=30).json()
    else:
        items = package_data.get("items")
        if not isinstance(items, list) or not items:
            return None
        last_item = items[-1]
        items_url = last_item.get("@id")
        if not isinstance(items_url, str):
            return None
        items_response = requests.get(items_url, timeout=30).json()
        items_list = items_response.get("items", [])
        try:
            package_data = next(
                x["catalogEntry"] for x in reversed(items_list) if "catalogEntry" in x
            )
        except StopIteration:
            package_data = next(
                (
                    y["catalogEntry"]
                    for x in reversed(items_list)
                    for y in reversed(x.get("items", []))
                    if "catalogEntry" in y and "pre" not in y["catalogEntry"].get("version", "")
                ),
                None,
            )

    if package_data is None:
        return None

    return cast(NugetCatalogEntry, package_data)



import os
from concurrent.futures import ThreadPoolExecutor

from labels.advisories.images.database import DATABASE as IMAGES_DATABASE
from labels.advisories.roots import DATABASE as ROOTS_DATABASE
from labels.config.bugsnag import initialize_bugsnag
from labels.config.logger import LOGGER, configure_logger, modify_logger_level
from labels.core.source_dispatcher import resolve_sbom_source
from labels.format import format_sbom
from labels.internal.file_resolver.container_image import ContainerImage
from labels.model.core import SbomConfig
from labels.pkg.cataloger.complete import complete_package
from labels.pkg.operations.package_operation import package_operations_factory
from labels.sources.directory_source import Directory


def initialize_scan_environment(sbom_config: SbomConfig) -> None:
    configure_logger()
    initialize_bugsnag()

    if sbom_config.debug:
        modify_logger_level()

    ROOTS_DATABASE.initialize()
    IMAGES_DATABASE.initialize()


def execute_labels_scan(sbom_config: SbomConfig) -> None:
    try:
        initialize_scan_environment(sbom_config)

        main_sbom_resolver = resolve_sbom_source(sbom_config)
        LOGGER.info(
            "📦 Generating SBOM from %s: %s",
            sbom_config.source_type.value,
            sbom_config.source,
        )

        packages, relationships = gather_packages_and_relationships(main_sbom_resolver)

        LOGGER.info("📦 Preparing %s report", sbom_config.output_format)
        format_sbom(
            packages=packages,
            relationships=relationships,
            config=sbom_config,
            resolver=main_sbom_resolver,
        )
    except Exception:
        LOGGER.exception(
            "Error executing labels scan",
            extra={"execution_id": sbom_config.execution_id},
        )
        raise


def gather_packages_and_relationships(
    resolver: Directory | ContainerImage,
    max_workers: int = 32,
) -> tuple[list, list]:
    packages, relationships = package_operations_factory(resolver)

    worker_count = min(
        max_workers,
        (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else max_workers,
    )
    with ThreadPoolExecutor(max_workers=worker_count) as executor:
        LOGGER.info("📦 Gathering additional package information")
        packages = list(filter(None, executor.map(complete_package, packages)))

    return packages, relationships



import logging
from typing import TYPE_CHECKING

from labels.file.scope import Scope
from labels.utils.strings import format_exception

if TYPE_CHECKING:
    from collections.abc import ItemsView
from typing import cast

from pydantic import ValidationError

from labels.artifact.relationship import Relationship
from labels.file.location import Location
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.internal.collection.json import parse_json_with_tree_sitter
from labels.internal.collection.types import IndexedDict, ParsedValue
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment
from labels.pkg.cataloger.python.package import package_url

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int, *, is_dev: bool) -> Location:
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: IndexedDict[str, ParsedValue]) -> str:
    version: str = value["version"] if isinstance(value["version"], str) else ""
    return version.strip("=<>~^ ")


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue | None,
    *,
    is_dev: bool = False,
) -> list[Package]:
    if dependencies is None or not isinstance(dependencies, IndexedDict):
        return []

    packages = []

    items: ItemsView[str, ParsedValue] = dependencies.items()
    for name, value in items:
        if not isinstance(value, IndexedDict) or not isinstance(name, str):
            continue
        version = _get_version(value)
        if not name or not version:
            continue

        location = _get_location(reader.location, value.position.start.line, is_dev=is_dev)

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    p_url=package_url(
                        name=name,
                        version=version,
                        package=None,
                    ),
                    licenses=[],
                    is_dev=is_dev,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def parse_pipfile_lock_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: ParsedValue | None = content.get("default")
    dev_deps: ParsedValue | None = content.get("develop")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



from fnmatch import (
    fnmatch,
)

import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)

from labels.pkg.cataloger.generic.cataloger import (
    Request,
)
from labels.pkg.cataloger.java.archive_parser import (
    parse_java_archive,
)
from labels.pkg.cataloger.java.parse_android_apk import (
    parse_apk,
)
from labels.pkg.cataloger.java.parse_build_stb import (
    build_stb,
)
from labels.pkg.cataloger.java.parse_gradle import (
    parse_gradle,
)
from labels.pkg.cataloger.java.parse_gradle_kts import (
    parse_gradle_lockfile_kts,
)
from labels.pkg.cataloger.java.parse_gradle_lockfile import (
    parse_gradle_lockfile,
)
from labels.pkg.cataloger.java.parse_gradle_properties import (
    parse_gradle_properties,
)
from labels.pkg.cataloger.java.parse_pom_xml import (
    parse_pom_xml,
)


def on_next_java(  # mccabe: disable=MC0001
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            patterns = [
                (
                    (
                        "**/pom.xml",
                        "/pom.xml",
                        "pom.xml",
                    ),
                    parse_pom_xml,
                    "java-parse-pom-xml",
                ),
                (
                    (
                        "**/*.jar",
                        "*.jar",
                        "**/*.war",
                        "*.war",
                        "**/*.ear",
                        "*.ear",
                        "**/*.par",
                        "*.par",
                        "**/*.sar",
                        "*.sar",
                        "**/*.nar",
                        "*.nar",
                        "**/*.jpi",
                        "*.jpi",
                        "**/*.hpi",
                        "*.hpi",
                        "**/*.lpkg",
                        "*.lpkg",
                    ),
                    parse_java_archive,
                    "java-archive-parse",
                ),
                (
                    (
                        "gradle.lockfile*",
                        "**/gradle.lockfile*",
                    ),
                    parse_gradle_lockfile,
                    "java-parse-gradle-lock",
                ),
                (
                    (
                        "**/*.gradle",
                        "/*.gradle",
                        "*.gradle",
                    ),
                    parse_gradle,
                    "java-parse-gradle-lock",
                ),
                (
                    (
                        "**/*.gradle.kts",
                        "*.gradle.kts",
                    ),
                    parse_gradle_lockfile_kts,
                    "java-parse-gradle-kts",
                ),
                (
                    (
                        "**/*.apk",
                        "*.apk",
                    ),
                    parse_apk,
                    "java-parse-apk",
                ),
                (
                    (
                        "**/build.sbt",
                        "build.sbt",
                    ),
                    build_stb,
                    "java-parse-build-stb",
                ),
                (
                    (
                        "**/gradle-wrapper.properties",
                        "/gradle-wrapper.properties",
                        "gradle-wrapper.properties",
                    ),
                    parse_gradle_properties,
                    "java-parse-gradle-properties",
                ),
            ]
            try:
                for pattern, parser, parser_name in patterns:
                    if any(fnmatch(value, x) for x in pattern):
                        observer.on_next(
                            Request(
                                real_path=value,
                                parser=parser,
                                parser_name=parser_name,
                            ),
                        )
                        break

            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from typing import (
    TypedDict,
)

from labels.internal.package_information.api_interface import (
    make_get,
)


class Release(TypedDict):
    version: str
    url: str
    inserted_at: str
    has_docs: bool


class Owner(TypedDict):
    url: str
    email: str
    username: str


class Downloads(TypedDict):
    all: int
    day: int
    recent: int
    week: int


class Meta(TypedDict):
    links: dict[str, str]
    description: str
    licenses: list[str]
    maintainers: list[str]


class Configs(TypedDict):
    erlang_mk: str
    mix_exs: str
    rebar_config: str


class PackageData(TypedDict):
    meta: Meta
    name: str
    url: str
    owners: list[Owner]
    inserted_at: str
    updated_at: str
    repository: str
    releases: list[Release]
    downloads: Downloads
    latest_version: str
    docs_html_url: str
    retirements: dict[str, str]
    configs: Configs
    html_url: str
    latest_stable_version: str


def get_hex_package(package_name: str) -> PackageData | None:
    url = f"https://hex.pm/api/packages/{package_name}"
    return make_get(url, timeout=30, headers={"Accept": "json"})



import logging
import sys

from bugsnag.handlers import (
    BugsnagHandler,
)

LOGGER_HANDLER = logging.StreamHandler()
LOGGER: logging.Logger = logging.getLogger()

FORMAT: str = "[%(levelname)s] %(message)s"

LOGGER_FORMATTER: logging.Formatter = logging.Formatter(FORMAT)

LOGGER_REMOTE_HANDLER = BugsnagHandler(extra_fields={"extra": ["extra"]})


def configure_logger() -> None:
    LOGGER_HANDLER.setStream(sys.stdout)
    LOGGER_HANDLER.setLevel(logging.INFO)
    LOGGER_HANDLER.setFormatter(LOGGER_FORMATTER)

    LOGGER.setLevel(logging.INFO)
    LOGGER.addHandler(LOGGER_HANDLER)

    LOGGER_REMOTE_HANDLER.setLevel(logging.ERROR)

    LOGGER.addFilter(LOGGER_REMOTE_HANDLER.leave_breadcrumbs)
    LOGGER.addHandler(LOGGER_REMOTE_HANDLER)


def modify_logger_level() -> None:
    LOGGER.setLevel(logging.DEBUG)
    LOGGER_HANDLER.setLevel(logging.DEBUG)



from labels.model.core import (
    Advisory,
    Package,
    PackageType,
)


def matches_namespace(package: Package, advisory: Advisory) -> bool:
    namespace_type = advisory.namespace.split(":")[1]
    if namespace_type == "distro":
        return package.type in {
            PackageType.DebPkg,
            PackageType.ApkPkg,
            PackageType.RpmPkg,
        }

    return False



import re

from labels.utils.licenses.licenses_index import (
    COMMON_LICENSES,
    INVERSE_LICENSES,
    sanitization_pattern,
)


def sanitize_license_string(license_str: str) -> str:
    sanitized = re.sub(r"\s+", " ", license_str.strip())
    sanitized = re.sub(r"[,\s]+", " ", sanitized)
    sanitized = re.sub(sanitization_pattern, "", sanitized)
    return re.sub(r"\s+", " ", sanitized).strip()


def find_license_by_pattern(sanitized_license: str) -> str | None:
    for identifier, (pattern, _) in COMMON_LICENSES.items():
        if re.search(pattern, sanitized_license, re.IGNORECASE):
            return identifier
    return None


def validate_licenses(licenses: list[str]) -> list[str]:
    found_licenses = set()
    for declared_license in licenses:
        sanitized_license = sanitize_license_string(declared_license)
        identifier = find_license_by_pattern(sanitized_license)
        if identifier:
            found_licenses.add(identifier)
        else:
            full_name = INVERSE_LICENSES.get(sanitized_license)
            if full_name:
                found_licenses.add(full_name)

    return list(found_licenses)



import logging

from packageurl import (
    PackageURL,
)
from pydantic import (
    BaseModel,
    ConfigDict,
    ValidationError,
)

from labels.advisories import (
    roots as advisories,
)
from labels.file.location import (
    Location,
)
from labels.internal.collection.types import ParsedValue
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class SwiftPackageManagerResolvedEntry(BaseModel):
    revision: str
    model_config = ConfigDict(frozen=True)


class CocoaPodfileLockEntry(BaseModel):
    checksum: str
    model_config = ConfigDict(frozen=True)


def new_cocoa_pods_package(
    name: str,
    version: str,
    hash_: ParsedValue,
    location: Location,
) -> Package | None:
    if not isinstance(hash_, str):
        return None
    try:
        return Package(
            name=name,
            version=version,
            p_url=cocoapods_package_url(name, version),
            locations=[location],
            type=PackageType.CocoapodsPkg,
            language=Language.SWIFT,
            metadata=CocoaPodfileLockEntry(checksum=hash_),
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None


def new_swift_package_manager_package(
    *,
    name: str,
    version: str,
    source_url: ParsedValue | None,
    revision: ParsedValue | None,
    location: Location,
) -> Package | None:
    try:
        return Package(
            name=name,
            version=version,
            p_url=swift_package_manager_package_url(name, version, source_url),
            locations=[location],
            type=PackageType.SwiftPkg,
            language=Language.SWIFT,
            metadata=SwiftPackageManagerResolvedEntry(revision=revision)
            if revision and isinstance(revision, str)
            else None,
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None


def cocoapods_package_url(
    name: str,
    version: str,
) -> str:
    return PackageURL("cocoapods", "", name, version, None, "").to_string()  # type: ignore


def swift_package_manager_package_url(
    name: str,
    version: str,
    source_url: ParsedValue | None,
) -> str:
    return PackageURL(  # type: ignore
        "swift",
        source_url.replace("https://", "", 1) if source_url and isinstance(source_url, str) else "",
        name,
        version,
        None,
        "",
    ).to_string()


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("swift", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)

    return package



import logging
from pathlib import Path
from typing import TypedDict, Unpack

import confuse

from labels.model.core import SbomConfig, SbomOutputFormat, SourceType
from labels.utils.exceptions import InvalidConfigFileError

LOGGER = logging.getLogger(__name__)


class ScanArgs(TypedDict):
    source: str
    format: str
    output: str
    docker_user: str | None
    docker_password: str | None
    aws_external_id: str | None
    aws_role: str | None
    config: bool
    debug: bool


def _build_config(config_path: str) -> confuse.Configuration:
    template = confuse.Configuration("labels", read=False)
    template.set_file(config_path)
    template.read(user=False, defaults=False)
    return template.get(
        confuse.Template(
            {
                "source": confuse.String(),
                "source_type": confuse.String(),
                "execution_id": confuse.String(),
                "exclude": confuse.Sequence(confuse.String()),
                "docker_credentials": confuse.Optional(
                    confuse.Template(
                        {
                            "username": confuse.Optional(confuse.String()),
                            "password": confuse.Optional(confuse.String()),
                        },
                    ),
                ),
                "aws_credentials": confuse.Optional(
                    confuse.Template(
                        {
                            "external_id": confuse.Optional(confuse.String()),
                            "role": confuse.Optional(confuse.String()),
                        },
                    ),
                ),
                "output": confuse.Template(
                    {
                        "name": confuse.String(),
                        "format": confuse.OneOf(
                            [_format.value for _format in SbomOutputFormat],
                        ),
                    },
                ),
                "debug": confuse.OneOf([True, False]),
            },
        ),
    )


def load(config_path: str) -> SbomConfig:
    config = _build_config(config_path)
    config_docker = config.pop("docker_credentials", {})
    config_aws = config.pop("aws_credentials", {})

    try:
        output = config.pop("output", None)

        sbom_config = SbomConfig(
            source=config.pop("source", None),
            source_type=config.pop("source_type", "dir"),
            output_format=output["format"],
            output=output["name"],
            exclude=config.pop("exclude", None) or (),
            docker_user=config_docker.pop("username", None),
            docker_password=config_docker.pop("password", None),
            aws_external_id=config_aws.pop("external_id", None),
            aws_role=config_aws.pop("role", None),
            execution_id=config.pop("execution_id", None),
            debug=config.pop("debug", False),
        )

        if config:
            unrecognized_keys = ", ".join(config)
            msg = (
                f"Some keys were not recognized: {unrecognized_keys}."
                " The analysis will be performed only using the supported keys"
                " and defaults."
            )
            LOGGER.warning(msg)
    except KeyError as exc:
        error_msg = f"Key: {exc.args[0]} is required"
        raise confuse.ConfigError(error_msg) from exc

    return sbom_config


def build_labels_config_from_args(arg: str, **kwargs: Unpack[ScanArgs]) -> SbomConfig:
    return SbomConfig(
        source=arg,
        source_type=SourceType.from_string(kwargs["source"]),
        execution_id=None,
        output_format=kwargs["format"],
        output=kwargs["output"],
        exclude=(),
        docker_user=kwargs["docker_user"],
        docker_password=kwargs["docker_password"],
        aws_external_id=kwargs["aws_external_id"],
        aws_role=kwargs["aws_role"],
        debug=kwargs["debug"],
    )


def build_labels_config_from_file(config_file_path: str) -> SbomConfig:
    if Path(config_file_path).is_file():
        if not config_file_path.endswith((".yaml", ".yml")):
            error_msg = "The configuration file must be a YAML format"
            raise InvalidConfigFileError(error_msg)

        return load(config_file_path)

    error_msg = f"The configuration file is not a valid file: {config_file_path}"
    raise InvalidConfigFileError(error_msg)



import re

import pefile
import semver
from packageurl import (
    PackageURL,
)
from pydantic import (
    BaseModel,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.context import (
    LOGGER,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)


class DotnetPortableExecutableEntry(BaseModel):
    assembly_version: str | None = None
    legal_copyright: str | None = None
    company_name: str | None = None
    product_name: str | None = None
    product_version: str | None = None
    comments: str | None = None
    internal_name: str | None = None


# Define the regular expressions
space_regex = re.compile(r"[\s]+")
number_regex = re.compile(r"\d")
version_punctuation_regex = re.compile(r"[.,]+")


def parse_version_resource(potable: pefile.PE) -> dict[str, str] | None:
    if not hasattr(potable, "VS_VERSIONINFO"):
        return None

    for idx, _ in enumerate(potable.VS_VERSIONINFO):
        if not hasattr(potable, "FileInfo") or len(potable.FileInfo) <= idx:
            continue

        stringtable_dict = process_file_info(
            potable.FileInfo[idx],  # type: ignore
        )
        if stringtable_dict:
            return stringtable_dict

    return None


def process_file_info(
    file_info: list[pefile.Structure],
) -> dict[str, str] | None:
    for entry in file_info:
        if not hasattr(entry, "StringTable"):
            continue

        stringtable_dict = process_string_table(entry.StringTable)
        if stringtable_dict:
            return stringtable_dict

    return None


def process_string_table(
    string_table: list[pefile.Structure],
) -> dict[str, str]:
    stringtable_dict = {}
    for st_entry in string_table:
        stringtable_dict["LangID"] = st_entry.LangID  # type: ignore
        for key, value in st_entry.entries.items():  # type: ignore
            stringtable_dict[key.decode("utf-8")] = value.decode("utf-8")
    return stringtable_dict


def is_microsoft(version_resources: dict[str, str]) -> bool:
    company_name = version_resources.get("CompanyName", "").lower()
    product_name = version_resources.get("ProductName", "").lower()

    return "microsoft" in company_name or "microsoft" in product_name


def space_normalize(value: str) -> str:
    # Trim leading and trailing whitespace
    value = value.strip()

    if value == "":
        return ""

    # Ensure valid UTF-8 text
    value = value.encode("utf-8", "replace").decode("utf-8")

    value = space_regex.sub(" ", value)

    # Remove other non-space, non-printable characters
    value = re.sub(r"[\x00-\x1f]", "", value)

    # Consolidate all space characters again
    value = space_regex.sub(" ", value)

    # Finally, remove any remaining surrounding whitespace
    return value.strip()


def find_name(version_resources: dict[str, str]) -> str:
    # Define the order of fields to check for the name
    name_fields = [
        "ProductName",
        "FileDescription",
        "InternalName",
        "OriginalFilename",
    ]

    # Check if the version resources are from Microsoft
    if is_microsoft(version_resources):
        # Adjust the order of fields for Microsoft-authored files
        name_fields = [
            "FileDescription",
            "InternalName",
            "OriginalFilename",
            "ProductName",
        ]

    # Iterate over the fields to find a non-empty, normalized name
    for field in name_fields:
        value = space_normalize(version_resources.get(field, ""))
        if value:
            return value

    # Return an empty string if no valid name is found
    return ""


def contains_number(string: str) -> bool:
    return any(char.isdigit() for char in string)


def extract_version(version: str) -> str:
    # Trim leading and trailing whitespace
    version = version.strip()

    out = ""

    # Split the version string into fields and iterate over them
    for index, char in enumerate(version.split()):
        # If the output already has a number but the current segment does not,
        # return the output
        if contains_number(out) and not contains_number(char):
            return out

        # Append the current field to the output
        if index == 0:
            out = char
        else:
            out += " " + char

    return out


def keep_greater_semantic_version(product_version: str, file_version: str) -> str:
    try:
        semantic_product_version = semver.VersionInfo.parse(product_version)
    except ValueError:
        LOGGER.debug(
            ("Unable to create semantic version from portable executable product version %s"),
            product_version,
        )
        return ""

    try:
        semantic_file_version = semver.VersionInfo.parse(file_version)
    except ValueError:
        LOGGER.debug(
            ("Unable to create semantic version from portable executable file version %s"),
            file_version,
        )
        return product_version

    # Make no choice when they are semantically equal
    if semantic_product_version == semantic_file_version:
        return ""

    if semantic_file_version > semantic_product_version:
        return file_version

    return product_version


def punctuation_count(string: str) -> int:
    # Find all matches of the punctuation regex and return their count
    return len(version_punctuation_regex.findall(string))


def find_version(version_resources: dict[str, str]) -> str:
    product_version = extract_version(version_resources.get("ProductVersion", ""))
    file_version = extract_version(version_resources.get("FileVersion", ""))

    semantic_version_compare_result = keep_greater_semantic_version(product_version, file_version)

    if semantic_version_compare_result:
        return semantic_version_compare_result

    product_version_detail = punctuation_count(product_version)
    file_version_detail = punctuation_count(file_version)

    if contains_number(product_version) and product_version_detail >= file_version_detail:
        return product_version

    if contains_number(file_version) and file_version_detail > 0:
        return file_version

    if contains_number(product_version):
        return product_version

    if contains_number(file_version):
        return file_version

    return product_version


def build_dot_net_package(
    version_resource: dict[str, str],
    reader: LocationReadCloser,
) -> Package | None:
    name = find_name(version_resource)
    if not name:
        LOGGER.debug(
            "Unable to find name for portable executable in file %s",
            reader.location.path(),
        )
        return None
    version = find_version(version_resource)
    if not version:
        LOGGER.debug(
            "Unable to find version for portable executable in file %s",
            reader.location.path(),
        )
        return None
    metadata = DotnetPortableExecutableEntry(
        assembly_version=version_resource.get("Assembly Version"),
        legal_copyright=version_resource.get("LegalCopyright"),
        comments=version_resource.get("Comments"),
        internal_name=version_resource.get("InternalName"),
        company_name=version_resource.get("CompanyName"),
        product_name=version_resource.get("ProductName"),
        product_version=version_resource.get("ProductVersion"),
    )

    return Package(
        name=name,
        version=version,
        locations=[reader.location],
        type=PackageType.DotnetPkg,
        language=Language.DOTNET,
        p_url=PackageURL(
            type="nuget",
            namespace="",
            name=name,
            version=version,
            qualifiers={},
            subpath="",
        ).to_string(),
        metadata=metadata,
        licenses=[],
    )


def parse_dotnet_portable_executable(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    if not reader.location.coordinates:
        return [], []
    try:
        pe_representation = pefile.PE(reader.location.coordinates.real_path, fast_load=False)
    except pefile.PEFormatError:
        return [], []

    version_resource = parse_version_resource(pe_representation)
    if not version_resource:
        LOGGER.debug(
            ("Unable to find version resource for portable executable in file %s"),
            reader.location.path(),
        )
        return [], []
    dotnet_package = build_dot_net_package(version_resource, reader)
    if not dotnet_package:
        LOGGER.debug(
            "Unable to build package for portable executable in file %s",
            reader.location.path(),
        )
        return [], []
    return [dotnet_package], []



import fnmatch
import os
import tarfile
import tempfile
from collections.abc import Generator
from contextlib import suppress
from pathlib import Path
from typing import TextIO, cast

from pydantic import BaseModel

from labels.file.location import Location, new_location_from_image
from labels.file.metadata import Metadata
from labels.file.resolver import FileReader, Resolver
from labels.file.type import get_type_from_tar_member
from labels.sources.docker import ImageContext, ImageMetadata


def is_in_dependency_dir(path: str, dependency_dirs: set[str]) -> bool:
    for dep in dependency_dirs:
        dep_norm = dep if dep.endswith(os.path.sep) else dep + os.path.sep
        if path.startswith(dep_norm):
            return "node_modules" in path
    return False


def _normalize_rel_root(root: str, base_path: str) -> str:
    rel_root = os.path.relpath(root, base_path)
    return os.path.sep if rel_root == "." else os.path.join(os.path.sep, rel_root)


class ContainerImage(Resolver, BaseModel):
    img: ImageMetadata
    context: ImageContext
    lazy: bool = True

    def _has_path_lazy(self, path: str) -> bool:
        success = []
        paths = [path, path.lstrip("/")]
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(os.path.join(self.context.layers_dir, layer_dir)) as tar:
                exists = False
                for _path in paths:
                    try:
                        tar.getmember(_path)
                        success.append(True)
                        exists = True
                        break
                    except KeyError:
                        success.append(False)
                if exists:
                    break

        return any(success)

    def _has_path(self, path: str) -> bool:
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        success = []
        path = path.lstrip(os.path.sep)
        for layer_id in layer_ids:
            p_file_path = os.path.join(self.context.full_extraction_dir, layer_id, path)
            if Path(p_file_path).exists():
                success.append(True)
                break

        return any(success)

    def has_path(self, path: str) -> bool:
        if self.lazy:
            return self._has_path_lazy(path)

        return self._has_path(path)

    def _search_path_lazy(self, path: str) -> list[Location]:
        # A path can be in multiple layers
        locations: list[Location] = []
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with (
                tarfile.open(os.path.join(self.context.layers_dir, layer_dir)) as tar,
                suppress(KeyError),
            ):
                member = tar.getmember(path)
                locations.append(
                    new_location_from_image(
                        member.name,
                        str(Path(layer_dir).parent),
                    ),
                )

        return locations

    def _search_path(self, path: str) -> list[Location]:
        if self.lazy:
            return self._search_path_lazy(path)

        locations: list[Location] = []
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        # A path can be in multiple layers
        for layer_id in layer_ids:
            p_file_path = os.path.join(
                self.context.full_extraction_dir,
                layer_id,
                path.lstrip(os.path.sep),
            )
            if Path(p_file_path).exists():
                locations.append(
                    new_location_from_image(
                        path,
                        layer_id,
                        p_file_path,
                    ),
                )
        return locations

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []

        for path in paths:
            if find_path := self._search_path(path.lstrip("/")):
                locations.extend(find_path)
            else:
                locations.extend(self._search_path(path))
        return locations

    def files_by_glob(self, *patterns: str) -> list[Location]:
        locations: list[Location] = []
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(os.path.join(self.context.layers_dir, layer_dir)) as tar:
                for member in tar.getmembers():
                    locations.extend(
                        new_location_from_image(
                            member.name,
                            str(Path(layer_dir).parent),
                            os.path.join(
                                self.context.full_extraction_dir,
                                member.name,
                            ),
                        )
                        for glob in patterns
                        if fnmatch.fnmatch(member.name, glob)
                    )
        return locations

    def _file_contents_by_location_lazy(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            if (
                location.coordinates
                and location.coordinates.file_system_id
                and location.coordinates.file_system_id not in layer_dir
            ):
                continue
            with (
                tarfile.open(os.path.join(self.context.layers_dir, layer_dir)) as tar,
                suppress(KeyError),
            ):
                if location.access_path:
                    member = tar.getmember(location.access_path)
                    if member.linkpath:
                        link_target_path = os.path.normpath(
                            os.path.join(
                                Path(member.name).parent,
                                member.linkname,
                            ),
                        )
                        member = tar.getmember(link_target_path)
                    temp_dir = tempfile.mkdtemp()
                    tar.extract(tar.getmember(member.name), temp_dir)
                    return cast(
                        TextIO,
                        (function_reader or open)(
                            os.path.join(temp_dir, member.name),
                            encoding="utf-8",
                            mode=mode or "r",
                        ),
                    )

        return None

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        if not self.lazy and location.coordinates and Path(location.coordinates.real_path).exists():
            return cast(
                TextIO,
                (function_reader or open)(
                    location.coordinates.real_path,
                    encoding="utf-8",
                    mode=mode or "r",
                ),
            )

        if self.lazy:
            return self._file_contents_by_location_lazy(
                location,
                function_reader=function_reader,
                mode=mode,
            )

        return None

    def file_metadata_by_location(
        self,
        location: Location,
    ) -> Metadata | None:
        if (
            location.coordinates is not None
            and location.coordinates.file_system_id is not None
            and (_layer_info := self.context.get_layer_info(location.coordinates.file_system_id))
        ):
            with tarfile.open(os.path.join(_layer_info.tar_full_path)) as tar:
                tar_member = tar.getmember(location.coordinates.real_path)

                return Metadata(
                    path=tar_member.name,
                    link_destination=tar_member.linkname,
                    user_id=tar_member.uid,
                    group_id=tar_member.gid,
                    type=get_type_from_tar_member(tar_member),
                    mime_type="",
                )

        return None

    def files_by_mime_type(self, _mime_type: str) -> list[Location]:
        raise NotImplementedError

    def relative_file_path(self, _: Location, _path: str) -> Location | None:
        files = self.files_by_path(_path)
        if not files:
            return None
        return files[0]

    def walk_file_lazy(self) -> Generator[str, None, None]:
        dependency_files = {"package.json", "package-lock.json", "yarn.lock"}
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(os.path.join(self.context.layers_dir, layer_dir)) as tar:
                members = tar.getmembers()
                dependency_dirs = set()
                for member in members:
                    if Path(member.name).name in dependency_files:
                        dependency_dirs.add(Path(member.name).parent)
                for member in members:
                    segments = member.name.split("/")
                    if "node_modules" in segments:
                        index = segments.index("node_modules")
                        parent_folder = "/".join(segments[:index])
                        if Path(parent_folder) in dependency_dirs:
                            continue
                    yield member.name

    def _walk_layer(
        self,
        current_layer_path: str,
        global_dependency_dirs: set[str],
    ) -> Generator[str, None, None]:
        dependency_files = {"package.json", "package-lock.json", "yarn.lock"}
        for root, dirs, files in os.walk(current_layer_path):
            rel_root = _normalize_rel_root(root, current_layer_path)
            if is_in_dependency_dir(rel_root, global_dependency_dirs):
                continue
            if any(dep in files for dep in dependency_files):
                global_dependency_dirs.add(rel_root)
                if "node_modules" in dirs:
                    dirs.remove("node_modules")

            for file in files:
                yield os.path.join(rel_root, file)

    def walk_file(self) -> Generator[str, None, None]:
        if self.lazy:
            yield from self.walk_file_lazy()
        else:
            layer_ids = [x["digest"] for x in self.context.manifest["layers"]]
            global_dependency_dirs: set[str] = set()

            for layer_id in layer_ids:
                current_layer_path = os.path.join(self.context.full_extraction_dir, layer_id)
                yield from self._walk_layer(current_layer_path, global_dependency_dirs)



import re
from collections.abc import Iterator
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from packageurl import (
    PackageURL,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.golang.package import (
    GolangModuleEntry,
)

GO_DIRECTIVE: re.Pattern[str] = re.compile(
    r"(?P<directive>require|replace) \(",
)
GO_MOD_DEP: re.Pattern[str] = re.compile(
    r"^\s+(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_REPLACE: re.Pattern[str] = re.compile(
    r"^\s+(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REP_DEP: re.Pattern[str] = re.compile(
    r"replace\s(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REQ_MOD_DEP: re.Pattern[str] = re.compile(
    r"require\s(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_VERSION: re.Pattern[str] = re.compile(
    r"\ngo (?P<major>\d)\.(?P<minor>\d+)(\.\d+)?\n",
)


def parse_go_sum_file(
    resolver: Resolver | None,
    reader: LocationReadCloser,
) -> dict[str, str] | None:
    out = {}

    if resolver is None or not reader.location.coordinates:
        return None

    go_sum_path = reader.location.coordinates.real_path.replace(".mod", ".sum")
    go_sum_location = resolver.relative_file_path(reader.location, go_sum_path)

    if go_sum_location is None:
        return None

    try:
        contents = resolver.file_contents_by_location(go_sum_location)
    except OSError:
        return None
    if not contents:
        return None
    # Read the contents of the file
    lines = contents.read().splitlines()

    for line in lines:
        parts = line.split(" ")
        if len(parts) < 3:
            continue

        name_version = f"{parts[0]} {parts[1]}"
        hash_value = parts[2]

        out[name_version] = hash_value

    return out


def package_url(module_name: str, module_version: str) -> str:
    fields = module_name.split("/")
    if not fields:
        return ""

    namespace = ""
    name = ""
    subpath = ""

    if len(fields) == 1:
        name = fields[0]
    elif len(fields) == 2:
        name = fields[1]
        namespace = fields[0]
    else:
        name = fields[2]
        namespace = "/".join(fields[:2])
        subpath = "/".join(fields[3:])

    return PackageURL(  # type: ignore
        type="golang",
        namespace=namespace,
        name=name,
        version=module_version,
        qualifiers=None,
        subpath=subpath,
    ).to_string()


def add_require(
    matched: re.Match[str],
    req_dict: dict[str, Package],
    line_number: int,
    parent_location: Location,
    sums: dict[str, str] | None,
) -> None:
    product: str = matched.group("product")
    version: str = matched.group("version")
    location = deepcopy(parent_location)
    location.scope = Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    sum_key = f"{product} v{version}"
    metadata = GolangModuleEntry(h1_digest=sums[sum_key]) if sums and sum_key in sums else None
    req_dict[product] = Package(
        name=product,
        version=version,
        type=PackageType.GoModulePkg,
        locations=[location],
        p_url=package_url(product, version),
        metadata=metadata,
        language=Language.GO,
        licenses=[],
        is_dev=False,
    )


def replace_req(
    req_dict: dict[str, Package],
    replace_list: list[tuple[re.Match[str], int]],
    parent_location: Location,
    sums: dict[str, str] | None,
) -> Iterator[Package]:
    for matched, line_number in replace_list:
        match_dict = cast(dict[str, str], matched.groupdict())
        old_pkg, old_version = match_dict["old_prod"], match_dict["old_ver"]
        repl_pkg, version = match_dict["new_prod"], match_dict["new_ver"]

        if old_pkg not in req_dict:
            continue

        if old_version and not version:
            version = req_dict[old_pkg].version

        if not version or (old_version and req_dict[old_pkg].version != old_version):
            continue

        location = deepcopy(parent_location)

        location.scope = Scope.PROD
        if location.coordinates:
            location.coordinates.line = line_number
            location.dependency_type = DependencyType.DIRECT

        sum_key = f"{repl_pkg} v{version}"
        metadata = GolangModuleEntry(h1_digest=sums[sum_key]) if sums and sum_key in sums else None

        req_dict[old_pkg] = Package(
            name=repl_pkg,
            version=version,
            type=PackageType.GoModulePkg,
            locations=[location],
            p_url=package_url(repl_pkg, version),
            metadata=metadata,
            language=Language.GO,
            licenses=[],
            is_dev=False,
        )

    return iter(req_dict.values())


def resolve_go_deps(
    content: str,
    location: Location,
    sums: dict[str, str] | None,
) -> Iterator[Package]:
    go_req_directive: str = ""
    replace_list: list[tuple[re.Match[str], int]] = []
    req_dict: dict[str, Package] = {}

    for line_number, line in enumerate(content.splitlines(), 1):
        if matched := GO_REQ_MOD_DEP.search(line):
            add_require(matched, req_dict, line_number, location, sums)
        elif replace := GO_REP_DEP.search(line):
            replace_list.append((replace, line_number))
        elif not go_req_directive:
            if directive := GO_DIRECTIVE.match(line):
                go_req_directive = directive.group("directive")
        elif go_req_directive == "replace":
            if replace := GO_REPLACE.search(line):
                replace_list.append((replace, line_number))
                continue
            go_req_directive = ""
        elif matched := GO_MOD_DEP.search(line):
            add_require(matched, req_dict, line_number, location, sums)
        else:
            go_req_directive = ""
    return replace_req(req_dict, replace_list, location, sums)


def parse_go_mod(
    resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    sums = parse_go_sum_file(resolver, reader)
    content = reader.read_closer.read()
    go_version = GO_VERSION.search(content)
    if not go_version:
        return [], []
    major = int(cast(dict[str, str], go_version.groupdict())["major"])
    minor = int(cast(dict[str, str], go_version.groupdict())["minor"])
    if major >= 2 or (major == 1 and minor >= 17):
        packages = list(resolve_go_deps(content, reader.location, sums))
    return packages, []



from collections.abc import Callable

import reactivex
from reactivex import (
    Observable,
)
from reactivex import (
    operators as ops,
)
from reactivex.scheduler import (
    ThreadPoolScheduler,
)

from labels.pkg.cataloger.cpp.cataloger import (
    on_next_cpp,
)
from labels.pkg.cataloger.dart.cataloger import (
    on_next_dart,
)
from labels.pkg.cataloger.dotnet.cataloger import (
    on_next_dotnet,
)
from labels.pkg.cataloger.elixir.cataloger import (
    on_next_elixir,
)
from labels.pkg.cataloger.generic.cataloger import (
    Request,
    on_next_db_file,
)
from labels.pkg.cataloger.github.cataloger import (
    on_next_github_action,
)
from labels.pkg.cataloger.golang.cataloger import (
    on_next_golang,
)
from labels.pkg.cataloger.java.cataloger import (
    on_next_java,
)
from labels.pkg.cataloger.javascript.cataloger import (
    on_next_javascript,
)
from labels.pkg.cataloger.php.cataloger import (
    on_next_php,
)
from labels.pkg.cataloger.python.cataloger import (
    on_next_python,
)
from labels.pkg.cataloger.redhat.cataloger import (
    on_next_redhat,
)
from labels.pkg.cataloger.ruby.cataloger import (
    on_next_ruby,
)
from labels.pkg.cataloger.rust.cataloger import (
    on_next_rust,
)
from labels.pkg.cataloger.swift.cataloger import (
    on_next_swift,
)


def handle_parser(
    scheduler: ThreadPoolScheduler,
) -> Callable[[Observable[str]], Observable[Request]]:
    def _apply_parsers(source: Observable[str]) -> Observable[Request]:
        return source.pipe(
            ops.flat_map(
                lambda item: reactivex.merge(  # type: ignore
                    (on_next_python(reactivex.just(item, scheduler))),
                    (on_next_db_file(reactivex.just(item, scheduler))),
                    (on_next_java(reactivex.just(item, scheduler))),
                    (on_next_javascript(reactivex.just(item, scheduler))),
                    (on_next_redhat(reactivex.just(item, scheduler))),
                    (on_next_dotnet(reactivex.just(item, scheduler))),
                    (on_next_rust(reactivex.just(item, scheduler))),
                    (on_next_ruby(reactivex.just(item, scheduler))),
                    (on_next_elixir(reactivex.just(item, scheduler))),
                    (on_next_php(reactivex.just(item, scheduler))),
                    (on_next_swift(reactivex.just(item, scheduler))),
                    (on_next_dart(reactivex.just(item, scheduler))),
                    (on_next_cpp(reactivex.just(item, scheduler))),
                    (on_next_golang(reactivex.just(item, scheduler))),
                    (on_next_github_action(reactivex.just(item, scheduler))),
                ),
            ),
        )

    return _apply_parsers



import logging
from typing import TYPE_CHECKING, cast

from pydantic import (
    ValidationError,
)

from labels.file.location import Location
from labels.model.core import (
    Advisory,
    HealthMetadata,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.alpine import (
    package as package_alpine,
)
from labels.pkg.cataloger.cpp import (
    package as package_cpp,
)
from labels.pkg.cataloger.dart import (
    package as package_dart,
)
from labels.pkg.cataloger.debian import (
    package as package_debian,
)
from labels.pkg.cataloger.dotnet import (
    package as package_dotnet,
)
from labels.pkg.cataloger.elixir import (
    package as package_elixir,
)
from labels.pkg.cataloger.github import (
    package as package_github,
)
from labels.pkg.cataloger.golang import (
    package as package_go,
)
from labels.pkg.cataloger.java import (
    package as package_java,
)
from labels.pkg.cataloger.javascript import (
    package as package_js,
)
from labels.pkg.cataloger.php import (
    package as package_php,
)
from labels.pkg.cataloger.python import (
    package as package_python,
)
from labels.pkg.cataloger.ruby import (
    package as package_ruby,
)
from labels.pkg.cataloger.rust import (
    package as package_rust,
)
from labels.pkg.cataloger.swift import (
    package as package_swift,
)
from labels.utils.strings import format_exception

if TYPE_CHECKING:
    from collections.abc import Callable

LOGGER = logging.getLogger(__name__)


def complete_package(package: Package) -> Package | None:
    completion_map: dict[PackageType, Callable[[Package], Package]] = {
        PackageType.NpmPkg: package_js.complete_package,
        PackageType.DartPubPkg: package_dart.complete_package,
        PackageType.DotnetPkg: package_dotnet.complete_package,
        PackageType.JavaPkg: package_java.complete_package,
        PackageType.PhpComposerPkg: package_php.complete_package,
        PackageType.PythonPkg: package_python.complete_package,
        PackageType.GemPkg: package_ruby.complete_package,
        PackageType.RustPkg: package_rust.complete_package,
        PackageType.DebPkg: package_debian.complete_package,
        PackageType.ApkPkg: package_alpine.complete_package,
        PackageType.GoModulePkg: package_go.complete_package,
        PackageType.HexPkg: package_elixir.complete_package,
        PackageType.ConanPkg: package_cpp.complete_package,
        PackageType.GithubActionPkg: package_github.complete_package,
        PackageType.SwiftPkg: package_swift.complete_package,
    }

    try:
        if package.type in completion_map:
            package = completion_map[package.type](package)
            package.model_validate(
                cast(
                    dict[
                        str,
                        str
                        | Language
                        | list[str]
                        | list[Location]
                        | PackageType
                        | list[Advisory]
                        | list[Package]
                        | HealthMetadata
                        | bool
                        | object
                        | None,
                    ],
                    package.__dict__,
                ),
            )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package completion. Required fields are missing "
            "or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": package.locations,
                    "package_type": package.type,
                },
            },
        )
        return None

    return package



import logging
import re
from contextlib import (
    suppress,
)

from bs4 import (
    BeautifulSoup,
)
from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.javascript.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

SCRIPT_DEP = re.compile(
    r"(?P<name>[^\s\/]*)(?P<separator>[-@\/])"
    r"(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))",
)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        return location.model_copy(
            update={"coordinates": location.coordinates.model_copy(update={"line": sourceline})},
        )
    return location


def parse_html_scripts(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    html = None
    with suppress(UnicodeError):
        try:
            html = BeautifulSoup(reader.read_closer, features="html.parser")
        except AssertionError:
            return [], []

    if not html:
        return [], []

    packages = []
    for script in html("script"):
        src_attribute = script.attrs.get("src")
        if not (src_attribute and src_attribute.endswith(".js")):
            continue

        matched = SCRIPT_DEP.search(src_attribute)

        if not matched:
            continue

        name = matched.group("name")
        version = matched.group("version")

        if not name or not version:
            continue

        location = _get_location(reader.location, script.sourceline)

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    licenses=[],
                    locations=[location],
                    language=Language.JAVASCRIPT,
                    type=PackageType.NpmPkg,
                    p_url=package_url(name, version),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



import logging
from copy import (
    deepcopy,
)

from bs4 import (
    BeautifulSoup,
)
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def parse_dotnet_pkgs_config(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    packages = []

    for pkg in root.find_all("package", recursive=True):
        name: str | None = pkg.get("id")
        version: str | None = pkg.get("version")

        if not name or not version:
            continue

        line = pkg.sourceline
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = line
            location.dependency_type = DependencyType.DIRECT

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.DOTNET,
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=name,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



import logging
import re
from copy import (
    deepcopy,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)

LOGGER = logging.getLogger(__name__)

GRADLE_DIST = re.compile("^distributionUrl=.+gradle-(?P<gradle_version>[^-]+)-.+")


def parse_gradle_properties(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    for line_no, raw_line in enumerate(reader.read_closer.readlines(), start=1):
        line = raw_line.strip()
        if (match := GRADLE_DIST.match(line)) and (version := match.group("gradle_version")):
            location = deepcopy(reader.location)
            if location.coordinates:
                location.coordinates.line = line_no

            packages.append(
                Package(
                    name="gradle",
                    version=version,
                    locations=[location],
                    language=Language.JAVA,
                    type=PackageType.JavaPkg,
                    p_url=package_url("gradle", version),
                    licenses=[],
                ),
            )

    return packages, []



import logging
from collections import (
    defaultdict,
)
from typing import (
    cast,
)

from cyclonedx.factory.license import (
    LicenseFactory,
)
from cyclonedx.model.bom import (
    Bom,
)
from cyclonedx.model.component import (
    Component,
    ComponentType,
)
from cyclonedx.model.license import (
    LicenseExpression,
)
from cyclonedx.model.tool import (
    Tool,
)
from packageurl import (
    PackageURL,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.resolver import Resolver
from labels.format.common import (
    set_namespace_version,
)
from labels.format.cyclone_dx.complete_file import (
    add_authors,
    add_component_properties,
    add_integrity,
    add_vulnerabilities,
)
from labels.format.cyclone_dx.output_handler import (
    format_cyclone_json,
    format_cyclone_xml,
)
from labels.model.core import (
    Package,
    SbomConfig,
)

LOGGER = logging.getLogger(__name__)


def pkg_to_component(package: Package) -> Component:
    lc_factory = LicenseFactory()
    licenses = [
        lc_factory.make_from_string(lic)
        for lic in package.licenses
        if not isinstance(lc_factory.make_from_string(lic), LicenseExpression)
    ]
    health_metadata = package.health_metadata
    return Component(
        type=ComponentType.LIBRARY,
        name=package.name,
        version=package.version,
        licenses=licenses,
        authors=add_authors(health_metadata) if health_metadata else [],
        bom_ref=f"{package.name}@{package.version}",
        purl=PackageURL.from_string(package.p_url),
        properties=add_component_properties(package),
        hashes=add_integrity(health_metadata) if health_metadata else [],
    )


def format_cyclonedx_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver,
) -> None:
    namespace, version = set_namespace_version(config=config, resolver=resolver)
    bom = Bom()
    bom.metadata.component = root_component = Component(
        name=namespace,
        type=ComponentType.APPLICATION,
        licenses=[],
        bom_ref="",
        version=version,
    )
    bom.metadata.tools.tools.add(Tool(vendor="Fluid Attacks", name="Fluid-Labels"))

    component_cache = {pkg: pkg_to_component(pkg) for pkg in packages}
    for component in component_cache.values():
        bom.components.add(component)
        bom.register_dependency(root_component, [component])

        package = next(
            pkg
            for pkg in packages
            if pkg.name == component.name and pkg.version == component.version
        )
        if package.advisories:
            vulnerabilities = add_vulnerabilities(package)
            for vulnerability in vulnerabilities:
                bom.vulnerabilities.add(vulnerability)

    dependency_map: dict[Component, list[Component]] = defaultdict(list)
    for relationship in relationships:
        to_pkg = component_cache.get(
            cast(Package, relationship.to_),
            pkg_to_component(cast(Package, relationship.to_)),
        )
        from_pkg = component_cache.get(
            cast(Package, relationship.from_),
            pkg_to_component(cast(Package, relationship.from_)),
        )
        dependency_map[to_pkg].append(from_pkg)

    for ref, depends_on_list in dependency_map.items():
        bom.register_dependency(ref, depends_on_list)

    match config.output_format:
        case "cyclonedx-json":
            format_cyclone_json(bom, config.output)
        case "cyclonedx-xml":
            format_cyclone_xml(bom, config.output)
